# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
This module contains custom Bodywork exceptions to be used throughout
all modules and tests. All Bodywork specific exceptions should be kept
in this module to make them easier to locate when importing the package
externally.
"""
from pathlib import Path
from typing import Iterable, Sequence

from kubernetes.client import V1Job

from .constants import BODYWORK_VERSION, BODYWORK_CONFIG_VERSION


class BodyworkJobFailure(Exception):
    def __init__(self, failed_jobs: Iterable[V1Job]):
        failed_jobs_msg = [
            f"job={job.metadata.name} in namespace={job.metadata.namespace}"
            for job in failed_jobs
        ]
        msg = f'{"; ".join(failed_jobs_msg)} have failed'
        super().__init__(msg)


class BodyworkConfigError(Exception):
    pass


class BodyworkConfigFileExistsError(BodyworkConfigError):
    def __init__(self, config_file_path: Path):
        msg = f"No config file found at {config_file_path}"
        super().__init__(msg)


class BodyworkConfigParsingError(BodyworkConfigError):
    def __init__(self, config_file_path: Path):
        msg = f"Cannot parse YAML from {config_file_path}"
        super().__init__(msg)


class BodyworkConfigMissingSectionError(BodyworkConfigError):
    def __init__(self, missing_sections: Sequence[str]):
        msg = f'Bodywork config file missing sections: {", ".join(missing_sections)}'
        super().__init__(msg)


class BodyworkConfigValidationError(BodyworkConfigError):
    def __init__(self, missing_params: Sequence[str]):
        self.missing_params = missing_params
        msg = (
            f"Bodywork config missing or invalid parameters: "
            f'{", ".join(missing_params)}'
        )
        super().__init__(msg)


class BodyworkConfigVersionMismatchError(BodyworkConfigError):
    def __init__(self, version: str):
        msg = (
            f"Bodywork config file has schema version {version}, when Bodywork "
            f"version {BODYWORK_VERSION} requires schema version "
            f"{BODYWORK_CONFIG_VERSION}"
        )
        super().__init__(msg)


class BodyworkWorkflowExecutionError(Exception):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class BodyworkStageFailure(Exception):
    def __init__(self, stage_name: str, info: str):
        msg = f"Stage {stage_name} failed - {info}"
        super().__init__(msg)


class BodyworkNamespaceError(Exception):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class BodyworkDockerImageError(Exception):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class BodyworkGitError(Exception):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
High-level interface to the Kubernetes APIs as used to create and manage
Bodywork service deployment stages.
"""
from datetime import datetime
from enum import Enum
from time import sleep, time
from typing import Dict, Iterable, List, Optional, Any

from kubernetes import client as k8s

from ..constants import BODYWORK_DOCKER_IMAGE, BODYWORK_STAGES_SERVICE_ACCOUNT
from .utils import make_valid_k8s_name


class DeploymentStatus(Enum):
    """Possible states of a k8s deployment."""

    COMPLETE = "complete"
    PROGRESSING = "progressing"


def configure_service_stage_deployment(
    namespace: str,
    stage_name: str,
    project_name: str,
    project_repo_url: str,
    git_commit_hash: str,
    project_repo_branch: str = "master",
    image: str = BODYWORK_DOCKER_IMAGE,
    replicas: int = 1,
    port: int = 80,
    container_env_vars: Optional[List[k8s.V1EnvVar]] = None,
    cpu_request: Optional[float] = None,
    memory_request: Optional[int] = None,
    seconds_to_be_ready_before_completing: int = 30,
) -> k8s.V1Deployment:
    """Configure a Bodywork service stage k8s deployment.

    :param namespace: The k8s namespace to target deployment.
    :param stage_name: The name of the Bodywork project stage that
        will need to be executed.
    :param project_name: The name of the Bodywork project that the stage
        belongs to.
    :param project_repo_url: The URL for the Bodywork project Git
        repository.
    :param git_commit_hash: The git commit hash of this Bodywork project.
    :param project_repo_branch: The Bodywork project Git repository
        branch to use, defaults to 'master'.
    :param image: Docker image to use for running the stage within,
        defaults to BODYWORK_DOCKER_IMAGE.
    :param replicas: Number of containers to create as part of this
        deployment, defaults to 1.
    :param port: The port to open when exposing the service, defaults
        to 80.
    :param container_env_vars: Optional list of environment variables
        (e.g. secrets) to set in the container, defaults to None.
    :param cpu_request: CPU resource to request from a node, expressed
        as a decimal number, defaults to None.
    :param memory_request: Memory resource to request from a node, expressed
        as an integer number of megabytes, defaults to None.
    :param seconds_to_be_ready_before_completing: Time (in seconds) that
        the deployment must be observed as being 'ready', before its
        status is moved to complete. Defaults to 30s.
    :return: A configured k8s deployment object.

    """
    service_name = make_valid_k8s_name(stage_name)
    container_resources = k8s.V1ResourceRequirements(
        requests={
            "cpu": f"{cpu_request}" if cpu_request else None,
            "memory": f"{memory_request}M" if memory_request else None,
        }
    )
    container = k8s.V1Container(
        name="bodywork",
        image=image,
        image_pull_policy="Always",
        resources=container_resources,
        env=container_env_vars,
        command=["bodywork", "stage"],
        args=[project_repo_url, project_repo_branch, stage_name],
    )
    pod_spec = k8s.V1PodSpec(
        service_account_name=BODYWORK_STAGES_SERVICE_ACCOUNT,
        containers=[container],
        restart_policy="Always",
    )
    pod_template_spec = k8s.V1PodTemplateSpec(
        metadata=k8s.V1ObjectMeta(
            labels={
                "app": "bodywork",
                "stage": service_name,
                "deployment-name": project_name,
                "git-commit-hash": git_commit_hash,
            },
            annotations={"last-updated": datetime.now().isoformat()},
        ),
        spec=pod_spec,
    )
    deployment_spec = k8s.V1DeploymentSpec(
        replicas=replicas,
        template=pod_template_spec,
        selector={"matchLabels": {"stage": service_name}},
        revision_history_limit=0,
        min_ready_seconds=seconds_to_be_ready_before_completing,
    )
    deployment_metadata = k8s.V1ObjectMeta(
        namespace=namespace,
        name=service_name,
        annotations={"port": str(port)},
        labels={
            "app": "bodywork",
            "stage": service_name,
            "deployment-name": project_name,
            "git-commit-hash": git_commit_hash,
        },
    )
    deployment = k8s.V1Deployment(metadata=deployment_metadata, spec=deployment_spec)
    return deployment


def create_deployment(deployment: k8s.V1Deployment) -> None:
    """Create a deployment on a k8s cluster.

    :param deployment: A configured deployment object.
    """
    k8s.AppsV1Api().create_namespaced_deployment(
        body=deployment, namespace=deployment.metadata.namespace
    )


def is_existing_deployment(namespace: str, name: str) -> bool:
    """Determine if the deployment already exists within the namespace.

    :param namespace: Namespace in which to look for deployment.
    :param name: Name of deployment in namespace.
    :return: Boolean flag for the deployment within the namespace.
    """
    existing_deployments = k8s.AppsV1Api().list_namespaced_deployment(
        namespace=namespace
    )
    existing_deployment_names = [
        deployment.metadata.name for deployment in existing_deployments.items
    ]
    return True if name in existing_deployment_names else False


def update_deployment(deployment: k8s.V1Deployment) -> None:
    """Update a deployment on a k8s cluster.

    :param deployment: A configured deployment object.
    """
    k8s.AppsV1Api().patch_namespaced_deployment(
        body=deployment,
        name=deployment.metadata.name,
        namespace=deployment.metadata.namespace,
    )


def rollback_deployment(deployment: k8s.V1Deployment) -> None:
    """Rollback a deployment to its previous version.

    The Kubernetes API has no dedicated endpoint for managing rollbacks.
    This function was implemented by reverse-engineering the API calls
    made by the equivalent kubectl command,`kubectl rollout undo ...`.

    :param deployment: A configured deployment object.
    """
    name = deployment.metadata.name
    namespace = deployment.metadata.namespace

    associated_replica_sets = k8s.AppsV1Api().list_namespaced_replica_set(
        namespace=namespace,
        label_selector=(f'app=bodywork,stage={deployment.metadata.labels["stage"]}'),
    )

    revision_ordered_replica_sets = sorted(
        associated_replica_sets.items,
        key=lambda e: e.metadata.annotations["deployment.kubernetes.io/revision"],
        reverse=True,
    )

    rollback_replica_set = (
        revision_ordered_replica_sets[0]
        if len(revision_ordered_replica_sets) == 1
        else revision_ordered_replica_sets[1]
    )

    rollback_revision_number = rollback_replica_set.metadata.annotations[
        "deployment.kubernetes.io/revision"
    ]

    is_new_deployment = (
        rollback_revision_number == "1" and len(revision_ordered_replica_sets) == 1
    )

    if is_new_deployment:
        delete_deployment(namespace, name)
    else:
        patch = [
            {
                "op": "replace",
                "path": "/spec/template",
                "value": rollback_replica_set.spec.template,
            },
            {
                "op": "replace",
                "path": "/metadata/annotations",
                "value": {
                    "deployment.kubernetes.io/revision": rollback_revision_number,
                    **deployment.metadata.annotations,
                },
            },
        ]
        k8s.AppsV1Api().patch_namespaced_deployment(
            body=patch, name=name, namespace=namespace
        )


def delete_all_namespace_deployments(namespace: str) -> None:
    """Delete all deployments within a k8s namespace.

    :param namespace: Namespace in which to look for deployments.
    """
    existing_deployments = k8s.AppsV1Api().list_namespaced_deployment(
        namespace=namespace
    )
    existing_deployment_names = [
        deployment.metadata.name for deployment in existing_deployments.items
    ]
    for name in existing_deployment_names:
        delete_deployment(namespace, name)


def delete_deployment(namespace: str, name: str) -> None:
    """Delete a deployment on a k8s cluster.

    :param namespace: Namespace in which to look for deployment.
    :param name: Name of deployment in namespace.
    """
    k8s.AppsV1Api().delete_namespaced_deployment(
        name=name,
        namespace=namespace,
        body=k8s.V1DeleteOptions(propagation_policy="Background"),
    )


def _get_deployment_status(deployment: k8s.V1Deployment) -> DeploymentStatus:
    """Get the latest status of a deployment created on a k8s cluster.

    :param deployment: A configured job object.
    :raises RuntimeError: If the deployment cannot be found or the status
        cannot be identified.
    :return: The current status of the deployment.
    """
    name = deployment.metadata.name
    namespace = deployment.metadata.namespace
    try:
        k8s_deployment_query = k8s.AppsV1Api().list_namespaced_deployment(
            namespace=namespace
        )
        k8s_deployment_data = [
            deployment
            for deployment in k8s_deployment_query.items
            if deployment.metadata.name == name
        ][0]
    except IndexError as e:
        msg = (
            f"cannot find deployment={deployment.metadata.name} in "
            f"namespace={deployment.metadata.namespace}"
        )
        raise RuntimeError(msg) from e

    if (
        k8s_deployment_data.status.available_replicas is not None
        and k8s_deployment_data.status.unavailable_replicas is None
    ):
        return DeploymentStatus.COMPLETE
    elif (
        k8s_deployment_data.status.available_replicas is None
        or k8s_deployment_data.status.unavailable_replicas > 0
    ):
        return DeploymentStatus.PROGRESSING
    else:
        msg = (
            f"cannot determine status for deployment={name} in "
            f"namespace={namespace}"
        )
        raise RuntimeError(msg)


def monitor_deployments_to_completion(
    deployments: Iterable[k8s.V1Deployment],
    timeout_seconds: int = 10,
    polling_freq_seconds: int = 1,
    wait_before_start_seconds: int = 5,
) -> bool:
    """Monitor deployment status until completion or timeout.

    :param deployments: The deployments to monitor.
    :param timeout_seconds: How long to keep monitoring status before
        calling a timeout, defaults to 10.
    :param polling_freq_seconds: Time (in seconds) between status
        polling, defaults to 1.
    :param wait_before_start_seconds: Time to wait before starting to
        monitor deployments - e.g. to allow deployments to be created.
    :raises TimeoutError: If the timeout limit is reached and the deployments
        are still marked as progressing.
    :return: True if all of the deployments are successful.
    """
    sleep(wait_before_start_seconds)
    start_time = time()
    deployments_status = [
        _get_deployment_status(deployment) for deployment in deployments
    ]
    while any(status is DeploymentStatus.PROGRESSING for status in deployments_status):
        sleep(polling_freq_seconds)
        if time() - start_time >= timeout_seconds:
            unsuccessful_deployments_msg = [
                (
                    f"deployment={deployment.metadata.name} in "
                    f"namespace={deployment.metadata.namespace}"
                )
                for deployment, status in zip(deployments, deployments_status)
                if status != DeploymentStatus.COMPLETE
            ]
            msg = (
                f'{"; ".join(unsuccessful_deployments_msg)} have yet to reach '
                f"status=complete after {timeout_seconds}s"
            )
            raise TimeoutError(msg)
        deployments_status = [
            _get_deployment_status(deployment) for deployment in deployments
        ]
    return True


def deployment_id(deployment_name: str, stage_name: str) -> str:
    """Return deployment ID implied by deployment and stage names.

    Args:
        deployment_name: The name given to the Bodywork deployment
            project.
        stage_name: The name of the stage that deployed a single
            service, within the Bodywork deployment project.

    Returns:
        Deployment ID string to use for locating services.
    """
    return f"{deployment_name}/{stage_name}"


def list_service_stage_deployments(
    namespace: Optional[str] = None,
    name: Optional[str] = None,
) -> Dict[str, Dict[str, Any]]:
    """Get all service deployments and their high-level info.

    :param namespace: Namespace in which to list services.
    :param name: Name of service.
    :return: Dict of deployments and their attributes.
    """
    label_selector = f"app=bodywork,deployment-name={name}" if name else "app=bodywork"
    if namespace:
        k8s_deployment_query = k8s.AppsV1Api().list_namespaced_deployment(
            namespace=namespace, label_selector=label_selector
        )
    else:
        k8s_deployment_query = k8s.AppsV1Api().list_deployment_for_all_namespaces(
            label_selector=label_selector
        )
    deployment_info = {}
    for deployment in k8s_deployment_query.items:
        exposed_as_cluster_service = is_exposed_as_cluster_service(
            deployment.metadata.namespace, deployment.metadata.name
        )
        id = deployment_id(
            deployment.metadata.labels["deployment-name"],
            deployment.metadata.labels["stage"],
        )
        deployment_info[id] = {
            "name": deployment.metadata.name,
            "namespace": deployment.metadata.namespace,
            "service_exposed": exposed_as_cluster_service,
            "service_url": (
                cluster_service_url(
                    deployment.metadata.namespace, deployment.metadata.name
                )
                if exposed_as_cluster_service
                else "none"
            ),
            "service_port": (
                deployment.metadata.annotations["port"]
                if exposed_as_cluster_service
                else "none"
            ),
            "available_replicas": (
                0
                if deployment.status.available_replicas is None
                else deployment.status.available_replicas
            ),
            "unavailable_replicas": (
                0
                if deployment.status.unavailable_replicas is None
                else deployment.status.unavailable_replicas
            ),
            "git_url": deployment.spec.template.spec.containers[0].args[0],
            "git_branch": deployment.spec.template.spec.containers[0].args[1],
            "git_commit_hash": deployment.metadata.labels.get("git-commit-hash", "NA"),
            "has_ingress": (
                has_ingress(deployment.metadata.namespace, deployment.metadata.name)
            ),
            "ingress_route": (
                ingress_route(deployment.metadata.namespace, deployment.metadata.name)
                if has_ingress(deployment.metadata.namespace, deployment.metadata.name)
                else "none"
            ),
        }

    return deployment_info


def cluster_service_url(namespace: str, deployment_name: str) -> str:
    """Standardised URL to a service deployment.

    :param namespace: Namespace in which the deployment exists.
    :param deployment_name: The deployment's name.
    :return: The internal URL to access the cluster service from within
        the cluster.
    """
    return f"http://{deployment_name}.{namespace}.svc.cluster.local"


def expose_deployment_as_cluster_service(deployment: k8s.V1Deployment) -> None:
    """Expose a deployment as a Kubernetes service.

    :param deployment: A configured deployment object.
    """
    namespace = deployment.metadata.namespace
    name = deployment.metadata.name
    pod_label_selector = deployment.spec.selector["matchLabels"]
    pod_port = int(deployment.metadata.annotations["port"])

    service_spec = k8s.V1ServiceSpec(
        type="ClusterIP",
        selector=pod_label_selector,
        ports=[k8s.V1ServicePort(port=pod_port, target_port=pod_port)],
    )
    service_metadata = k8s.V1ObjectMeta(
        namespace=namespace, name=name, labels={"app": "bodywork", "stage": name}
    )
    service = k8s.V1Service(metadata=service_metadata, spec=service_spec)
    k8s.CoreV1Api().create_namespaced_service(namespace=namespace, body=service)


def is_exposed_as_cluster_service(namespace: str, name: str) -> bool:
    """Is a deployment exposed as a cluster service.

    :param namespace: Namespace in which to look for services.
    :param name: The name of the service.
    """
    services = k8s.CoreV1Api().list_namespaced_service(namespace=namespace)
    service_names = [service.metadata.name for service in services.items]
    return True if name in service_names else False


def stop_exposing_cluster_service(namespace: str, name: str) -> None:
    """Delete a service associated with a deployment.

    :param namespace: Namespace in which exists the service to delete.
    :param name: The name of the service.
    """
    k8s.CoreV1Api().delete_namespaced_service(
        namespace=namespace, name=name, propagation_policy="Background"
    )


def ingress_route(namespace: str, deployment_name: str) -> str:
    """Standardised route to a service deployment.

    :param namespace: Namespace in which the deployment exists.
    :param deployment_name: The deployment's name.
    :return: A route to use for ingress to the service deployment.
    """
    return f"/{namespace}/{deployment_name}"


def create_deployment_ingress(deployment: k8s.V1Deployment) -> None:
    """Create an ingress to a service backed by a deployment.

    :param deployment: A configured deployment object.
    """
    namespace = deployment.metadata.namespace
    name = deployment.metadata.name
    pod_port = int(deployment.metadata.annotations["port"])

    ingress_path = f"{ingress_route(namespace, name)}(/|$)(.*)"

    ingress_spec = k8s.V1IngressSpec(
        rules=[
            k8s.V1IngressRule(
                http=k8s.V1HTTPIngressRuleValue(
                    paths=[
                        k8s.V1HTTPIngressPath(
                            path=ingress_path,
                            path_type="Exact",
                            backend=k8s.V1IngressBackend(
                                service=k8s.V1IngressServiceBackend(
                                    name=name,
                                    port=k8s.V1ServiceBackendPort(number=pod_port),
                                )
                            ),
                        )
                    ]
                )
            )
        ]
    )

    ingress_metadata = k8s.V1ObjectMeta(
        namespace=namespace,
        name=name,
        annotations={
            "kubernetes.io/ingress.class": "nginx",
            "nginx.ingress.kubernetes.io/rewrite-target": "/$2",
        },
        labels={"app": "bodywork", "stage": name},
    )

    ingress = k8s.V1Ingress(metadata=ingress_metadata, spec=ingress_spec)

    k8s.NetworkingV1Api().create_namespaced_ingress(namespace=namespace, body=ingress)


def delete_deployment_ingress(namespace: str, name: str) -> None:
    """Delete an ingress to a service backed by a deployment.

    :param namespace: Namespace in which exists the ingress to delete.
    :param name: The name of the ingress.
    """
    k8s.NetworkingV1Api().delete_namespaced_ingress(
        namespace=namespace, name=name, propagation_policy="Background"
    )


def has_ingress(namespace: str, name: str) -> bool:
    """Does a service backed by a deployment have an ingress?

    :param namespace: Namespace in which to look for ingress resources.
    :param name: The name of the ingress.
    """
    ingresses = k8s.NetworkingV1Api().list_namespaced_ingress(
        namespace=namespace, label_selector="app=bodywork"
    )
    ingress_names = [ingress.metadata.name for ingress in ingresses.items]
    return True if name in ingress_names else False



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
High-level interface to the Kubernetes jobs and cronjobs APIs, as used
to create and manage cronjobs that execute Bodywork project workflows.
"""
import os
import random
from datetime import datetime
from typing import Dict, Union, Optional, List
from kubernetes import client as k8s

from ..constants import (
    BODYWORK_DOCKER_IMAGE,
    BODYWORK_WORKFLOW_SERVICE_ACCOUNT,
    BODYWORK_WORKFLOW_JOB_TIME_TO_LIVE,
)
from .utils import make_valid_k8s_name


def configure_workflow_job(
    namespace: str,
    project_repo_url: str,
    project_repo_branch: str = "master",
    retries: int = 2,
    image: str = BODYWORK_DOCKER_IMAGE,
    job_name: Optional[str] = None,
    container_env_vars: Optional[List[k8s.V1EnvVar]] = None,
) -> k8s.V1Job:
    """Configure a Bodywork workflow execution job.

    :param namespace: The namespace to deploy the job to.
    :param project_repo_url: The URL for the Bodywork project Git
        repository.
    :param project_repo_branch: The Bodywork project Git repository
        branch to use, defaults to 'master'.
    :param retries: Number of times to retry running the stage to
        completion (if necessary), defaults to 2.
    :param image: Docker image to use for running the stage within,
        defaults to BODYWORK_DOCKER_IMAGE.
    :param job_name: Set the job name.
    :param container_env_vars: List of k8s environment variables to add.
    :return: A configured k8s job object.
    """

    container = k8s.V1Container(
        name="bodywork",
        image=image,
        image_pull_policy="Always",
        env=container_env_vars,
        command=[
            "bodywork",
            "create",
            "deployment",
        ],
        args=[project_repo_url, project_repo_branch],
    )
    pod_spec = k8s.V1PodSpec(
        service_account_name=BODYWORK_WORKFLOW_SERVICE_ACCOUNT,
        containers=[container],
        restart_policy="Never",
    )
    pod_template_spec = k8s.V1PodTemplateSpec(spec=pod_spec)
    job_spec = k8s.V1JobSpec(
        template=pod_template_spec,
        completions=1,
        backoff_limit=retries,
        ttl_seconds_after_finished=BODYWORK_WORKFLOW_JOB_TIME_TO_LIVE,
    )
    if not job_name:
        job_name = _create_job_name(project_repo_url, project_repo_branch)
    job = k8s.V1Job(
        metadata=k8s.V1ObjectMeta(
            name=make_valid_k8s_name(job_name),
            namespace=namespace,
            labels={"app": "bodywork", "deployment-name": job_name},
        ),
        spec=job_spec,
    )
    return job


def _create_job_name(project_repo_url: str, project_repo_branch: str) -> str:
    """Create unique job name.

    :param project_repo_url: The URL for the Bodywork project Git
        repository.
    :param project_repo_branch: The Bodywork project Git repository
        branch being used.
    """
    repo_name = os.path.splitext(os.path.basename(project_repo_url))[0]
    return f"{repo_name}-{project_repo_branch}-{random.randint(0,999999)}"


def create_workflow_job(job: k8s.V1Job) -> None:
    """Create a workflow execution job.

    :param job: A configured job object.
    """
    k8s.BatchV1Api().create_namespaced_job(body=job, namespace=job.metadata.namespace)


def configure_workflow_cronjob(
    cron_schedule: str,
    namespace: str,
    job_name: str,
    project_repo_url: str,
    project_repo_branch: str = "master",
    retries: int = 2,
    successful_jobs_history_limit: int = 1,
    failed_jobs_history_limit: int = 1,
    image: str = BODYWORK_DOCKER_IMAGE,
    env_vars: Optional[List[k8s.V1EnvVar]] = None,
) -> k8s.V1beta1CronJob:
    """Configure a Bodywork workflow cronjob.

    A cronjob is a k8s job that is executed on a cron-like schedule. In
    this particular instance, the job will execute the `run_workflow`
    function that will orchestrate the required jobs and deployments.

    :param cron_schedule: A valid cron schedule definition.
    :param job_name: The name to give the cronjob.
    :param namespace: Namespace to create cronjob in.
    :param project_repo_url: The URL for the Bodywork project Git
        repository.
    :param project_repo_branch: The Bodywork project Git repository
        branch to use, defaults to 'master'.
    :param retries: Number of times to retry running the stage to
        completion (if necessary), defaults to 2.
    :param successful_jobs_history_limit: The number of successful job
        runs (pods) to keep, defaults to 1.
    :param failed_jobs_history_limit: The number of unsuccessful job
        runs (pods) to keep, defaults to 1.
    :param image: Docker image to use for running the stage within,
        defaults to BODYWORK_DOCKER_IMAGE.
    :param env_vars: List of k8s environment variables to add.
    :return: A configured k8s cronjob object.
    """
    job = configure_workflow_job(
        namespace=namespace,
        project_repo_url=project_repo_url,
        project_repo_branch=project_repo_branch,
        retries=retries,
        image=image,
        job_name=job_name,
        container_env_vars=env_vars,
    )
    job_template = k8s.V1beta1JobTemplateSpec(metadata=job.metadata, spec=job.spec)
    cronjob_spec = k8s.V1beta1CronJobSpec(
        schedule=cron_schedule,
        successful_jobs_history_limit=successful_jobs_history_limit,
        failed_jobs_history_limit=failed_jobs_history_limit,
        job_template=job_template,
    )
    cronjob = k8s.V1beta1CronJob(metadata=job.metadata, spec=cronjob_spec)
    return cronjob


def create_workflow_cronjob(cron_job: k8s.V1Job) -> None:
    """Create a cron-job on a k8s cluster.

    :param cron_job: A configured cron-job object.
    """
    k8s.BatchV1beta1Api().create_namespaced_cron_job(
        body=cron_job, namespace=cron_job.metadata.namespace
    )


def update_workflow_cronjob(
    namespace: str,
    name: str,
    schedule: Optional[str] = None,
    project_repo_url: Optional[str] = None,
    project_repo_branch: Optional[str] = None,
    retries: Optional[int] = None,
    successful_jobs_history_limit: Optional[int] = None,
    failed_jobs_history_limit: Optional[int] = None,
) -> None:
    """Update a Bodywork workflow cronjob.

    :param name: The name  of the cronjob.
    :param namespace: Namespace the cronjob resides in.
    :param schedule: A valid cron schedule definition.
    :param project_repo_url: The URL for the Bodywork project Git
        repository.
    :param project_repo_branch: The Bodywork project Git repository
        branch to use, defaults to 'master'.
    :param retries: Number of times to retry running the stage to
        completion (if necessary), defaults to 2.
    :param successful_jobs_history_limit: The number of successful job
        runs (pods) to keep, defaults to 1.
    :param failed_jobs_history_limit: The number of unsuccessful job
        runs (pods) to keep, defaults to 1.
    """

    if not schedule:
        schedule = (
            k8s.BatchV1beta1Api()
            .read_namespaced_cron_job(name, namespace)
            .spec.schedule
        )

    if project_repo_url and project_repo_branch:
        pod_spec = k8s.V1PodSpec(
            containers=[
                k8s.V1Container(
                    name="bodywork",
                    command=[
                        "bodywork",
                        "create",
                        "deployment",
                    ],
                    args=[project_repo_url, project_repo_branch],
                )
            ],
        )
        job = k8s.V1Job(
            spec=k8s.V1JobSpec(
                template=k8s.V1PodTemplateSpec(spec=pod_spec),
                backoff_limit=retries,
            )
        )
        job_template = k8s.V1beta1JobTemplateSpec(spec=job.spec)
    else:
        job_template = k8s.V1beta1JobTemplateSpec()

    cronjob = k8s.V1beta1CronJob(
        spec=k8s.V1beta1CronJobSpec(
            job_template=job_template,
            schedule=schedule,
            successful_jobs_history_limit=successful_jobs_history_limit,
            failed_jobs_history_limit=failed_jobs_history_limit,
        )
    )

    k8s.BatchV1beta1Api().patch_namespaced_cron_job(name, namespace, cronjob)


def delete_workflow_cronjob(namespace: str, name: str) -> None:
    """Delete a cron-job on a k8s cluster.

    :param namespace: Namespace in which to look for the cronjob to
        delete.
    :param name: The name of the cronjob to be deleted.
    """
    k8s.BatchV1beta1Api().delete_namespaced_cron_job(
        name=name,
        namespace=namespace,
        body=k8s.V1DeleteOptions(propagation_policy="Background"),
    )


def list_workflow_cronjobs(namespace: str) -> Dict[str, Dict[str, str]]:
    """Get all cronjobs and their high-level info.

    :param namespace: Namespace in which to list cronjobs.
    """
    cronjobs = k8s.BatchV1beta1Api().list_namespaced_cron_job(namespace=namespace)
    cronjob_info = {
        cronjob.metadata.name: {
            "schedule": cronjob.spec.schedule,
            "last_scheduled_time": cronjob.status.last_schedule_time,
            "retries": cronjob.spec.job_template.spec.backoff_limit,
            "git_url": (
                cronjob.spec.job_template.spec.template.spec.containers[0].args[0]
            ),
            "git_branch": (
                cronjob.spec.job_template.spec.template.spec.containers[0].args[1]
            ),
        }
        for cronjob in cronjobs.items
    }
    return cronjob_info


def list_workflow_jobs(
    namespace: str, job_name: str
) -> Dict[str, Dict[str, Union[datetime, bool]]]:
    """Get historic workflow jobs.

    Get status information for workflow jobs owned by a job or cronjob.

    :param namespace: Namespace in which to list workflow jobs.
    :param job_name: Name of job that triggered workflow job.
    :return: Dictionary of workflow jobs each mapping to a dictionary of
        status information fields for the workflow.
    """
    workflow_jobs_query = k8s.BatchV1Api().list_namespaced_job(namespace=namespace)
    workflow_jobs_info = {
        workflow_job.metadata.name: {
            "start_time": workflow_job.status.start_time,
            "completion_time": workflow_job.status.completion_time,
            "active": True if workflow_job.status.active else False,
            "succeeded": True if workflow_job.status.succeeded else False,
            "failed": True if workflow_job.status.failed else False,
        }
        for workflow_job in workflow_jobs_query.items
        if workflow_job.metadata.name.startswith(job_name)
    }
    return workflow_jobs_info



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Kubernetes API sub-module.

This collection of functions utilise the low-level Kubernetes Python
client to perform Bodywork-specific tasks.
"""
from .auth import (
    cluster_role_binding_exists,
    cluster_role_exists,
    delete_cluster_role_binding,
    workflow_cluster_role_binding_name,
    load_kubernetes_config,
    service_account_exists,
    setup_stages_service_account,
    setup_workflow_service_accounts,
)
from .workflow_jobs import (
    configure_workflow_job,
    configure_workflow_cronjob,
    create_workflow_job,
    create_workflow_cronjob,
    delete_workflow_cronjob,
    list_workflow_cronjobs,
    list_workflow_jobs,
    update_workflow_cronjob,
)
from .batch_jobs import (
    JobStatus,
    configure_batch_stage_job,
    create_job,
    delete_job,
    monitor_jobs_to_completion,
)
from .namespaces import namespace_exists, create_namespace, delete_namespace
from .pod_logs import get_latest_pod_name, get_pod_logs
from .secrets import (
    configure_env_vars_from_secrets,
    secret_exists,
    create_secret,
    delete_secret,
    list_secrets,
    replicate_secrets_in_namespace,
    update_secret,
    Secret,
    create_complete_secret_name,
    create_ssh_key_secret_from_file,
    create_secret_env_variable,
)
from .deployments import (
    DeploymentStatus,
    configure_service_stage_deployment,
    create_deployment,
    deployment_id,
    is_existing_deployment,
    update_deployment,
    rollback_deployment,
    delete_deployment,
    delete_all_namespace_deployments,
    monitor_deployments_to_completion,
    list_service_stage_deployments,
    cluster_service_url,
    expose_deployment_as_cluster_service,
    is_exposed_as_cluster_service,
    stop_exposing_cluster_service,
    ingress_route,
    create_deployment_ingress,
    delete_deployment_ingress,
    has_ingress,
)
from .utils import (
    api_exception_msg,
    create_k8s_environment_variables,
    EnvVars,
    make_valid_k8s_name,
)


__all__ = [
    "cluster_role_binding_exists",
    "cluster_role_exists",
    "delete_cluster_role_binding",
    "workflow_cluster_role_binding_name",
    "load_kubernetes_config",
    "service_account_exists",
    "setup_stages_service_account",
    "setup_workflow_service_accounts",
    "configure_workflow_cronjob",
    "create_workflow_job",
    "create_workflow_cronjob",
    "delete_workflow_cronjob",
    "list_workflow_cronjobs",
    "list_workflow_jobs",
    "JobStatus",
    "configure_workflow_job",
    "configure_batch_stage_job",
    "create_job",
    "delete_job",
    "monitor_jobs_to_completion",
    "namespace_exists",
    "create_namespace",
    "delete_namespace",
    "get_latest_pod_name",
    "get_pod_logs",
    "configure_env_vars_from_secrets",
    "secret_exists",
    "create_secret",
    "delete_secret",
    "list_secrets",
    "DeploymentStatus",
    "configure_service_stage_deployment",
    "create_deployment",
    "deployment_id",
    "is_existing_deployment",
    "update_deployment",
    "rollback_deployment",
    "delete_deployment",
    "delete_all_namespace_deployments",
    "monitor_deployments_to_completion",
    "list_service_stage_deployments",
    "cluster_service_url",
    "expose_deployment_as_cluster_service",
    "is_exposed_as_cluster_service",
    "stop_exposing_cluster_service",
    "ingress_route",
    "create_deployment_ingress",
    "delete_deployment_ingress",
    "has_ingress",
    "api_exception_msg",
    "create_k8s_environment_variables",
    "EnvVars",
    "make_valid_k8s_name",
    "replicate_secrets_in_namespace",
    "update_secret",
    "Secret",
    "update_workflow_cronjob",
    "create_complete_secret_name",
    "create_ssh_key_secret_from_file",
    "create_secret_env_variable",
]



import sys
import traceback
import urllib3
import warnings
from datetime import datetime
from functools import wraps
from pathlib import Path
from time import sleep
from typing import Any, Callable, List, Optional

import kubernetes
from pkg_resources import get_distribution
from typer import Argument, Option, Typer

from bodywork.k8s.utils import make_valid_k8s_name
from ..config import BodyworkConfig
from .terminal import print_info, print_warn
from bodywork.cli.workflow_jobs import (
    create_workflow_job,
    create_workflow_cronjob,
    display_cronjobs,
    display_workflow_job_history,
    display_workflow_job_logs,
    delete_workflow_cronjob,
    delete_workflow_job,
    update_workflow_cronjob,
)
from .deployments import display_deployments, delete_deployment
from .secrets import (
    create_secret,
    delete_secret,
    display_secrets,
    parse_cli_secrets_strings,
    update_secret,
)
from .setup_namespace import (
    is_namespace_available_for_bodywork,
    setup_namespace_with_service_accounts_and_roles,
)
from .terminal import console
from ..exceptions import (
    BodyworkConfigFileExistsError,
    BodyworkConfigValidationError,
    BodyworkConfigMissingSectionError,
    BodyworkConfigParsingError,
    BodyworkWorkflowExecutionError,
)
from ..constants import BODYWORK_NAMESPACE, BODYWORK_DOCKER_IMAGE
from ..k8s import api_exception_msg, load_kubernetes_config
from ..stage_execution import run_stage
from bodywork.workflow_execution import run_workflow

warnings.simplefilter(action="ignore")

cli_app = Typer()

create = Typer()
cli_app.add_typer(create, name="create")

get = Typer()
cli_app.add_typer(get, name="get")

update = Typer()
cli_app.add_typer(update, name="update")

delete = Typer()
cli_app.add_typer(delete, name="delete")


def k8s_auth(func: Callable[..., None]) -> Callable[..., None]:
    """Decorator for handling k8s authentication for CLI commands.

    :param func: The inner function to wrap with k8s exception handling.
    :return: The original function wrapped by a function that handles
        k8s API exceptions.
    """
    try:
        load_kubernetes_config()
    except Exception as e:
        print_warn(f"Could not authenticate with active Kubernetes context. \n--> {e}")
    return func


def handle_k8s_exceptions(func: Callable[..., None]) -> Callable[..., None]:
    """Decorator for handling k8s API exceptions on the CLI.

    :param func: The inner function to wrap with k8s exception handling.
    :return: The original function wrapped by a function that handles
        k8s API exceptions.
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> None:
        try:
            func(*args, **kwargs)
        except kubernetes.client.rest.ApiException:
            e_type, e_value, e_tb = sys.exc_info()
            exception_origin = traceback.extract_tb(e_tb)[2].name
            print_warn(
                f"Kubernetes API error returned when called from {exception_origin} "
                f"within cli.{func.__name__}: {api_exception_msg(e_value)}"
            )
        except urllib3.exceptions.MaxRetryError:
            e_type, e_value, e_tb = sys.exc_info()
            exception_origin = traceback.extract_tb(e_tb)[2].name
            print_warn(
                f"Failed to connect to the Kubernetes API when called from "
                f"{exception_origin} within cli.{func.__name__}: {e_value}"
            )
        except kubernetes.config.ConfigException as e:
            print_warn(
                f"Cannot load authentication credentials from kubeconfig file when "
                f"calling cli.{func.__name__}: {e}"
            )

    return wrapper


@cli_app.command("validate")
def _validate_config(
    file: str = Option("bodywork.yaml"), check_files: bool = Option(False)
):
    file_path = Path(file)
    try:
        BodyworkConfig(file_path, check_files)
        print_info(f"--> {file_path} is a valid Bodywork config file.")
        sys.exit(0)
    except (
        BodyworkConfigFileExistsError,
        BodyworkConfigParsingError,
        BodyworkConfigMissingSectionError,
    ) as e:
        print_warn(f"--> {e}")
        sys.exit(1)
    except BodyworkConfigValidationError as e:
        print_warn(f"Missing or invalid parameters found in {file_path}:")
        missing_or_invalid_param_list = "\n* ".join(e.missing_params)
        print_warn(f"* {missing_or_invalid_param_list}")
        sys.exit(1)


@cli_app.command("version")
def _version():
    print_info(get_distribution("bodywork").version)
    sys.exit(0)


@cli_app.command("configure-cluster")
@handle_k8s_exceptions
@k8s_auth
def _configure_cluster():
    setup_namespace_with_service_accounts_and_roles(BODYWORK_NAMESPACE)
    sys.exit(0)


@cli_app.command("stage", hidden=True)
def _stage(
    git_url: str = Argument(...),
    git_branch: str = Argument(...),
    stage_name: str = Argument(...),
):
    try:
        run_stage(stage_name, git_url, git_branch)
        sys.exit(0)
    except Exception:
        sys.exit(1)


@cli_app.command("debug", hidden=True)
def _debug(seconds: int = Argument(600)) -> None:
    print_info(f"sleeping for {seconds}s")
    sleep(seconds)
    sys.exit(0)


@create.command("deployment")
@handle_k8s_exceptions
@k8s_auth
def _create_deployment(
    git_url: str = Argument(...),
    git_branch: str = Argument(...),
    asynchronous: bool = Option(False, "--async"),
    asynchronous_job_name: str = Option("", "--async-job-name"),
    ssh_key_path: str = Option("", "--ssh"),
    secrets_group: str = Option("", "--group", "--secrets-group"),
    image: Optional[str] = Option(None, "--bodywork-image"),
    retries: int = Option(1),
):
    if not is_namespace_available_for_bodywork(BODYWORK_NAMESPACE):
        print_warn(
            "Cluster has not been configured for Bodywork - "
            "running 'bodywork configure-cluster'."
        )
        setup_namespace_with_service_accounts_and_roles(
            BODYWORK_NAMESPACE
        )
    if not asynchronous:
        print_info("Using local workflow controller - retries inactive.")
        try:
            console.rule(
                f"[green]deploying[/green] [bold purple]{git_branch}[/bold purple] "
                f"[green]branch from[/green] [bold purple]{git_url}[/bold purple]",
                characters="=",
                style="green",
            )
            with console.status(
                "[purple]Bodywork deploying[/purple]", spinner="aesthetic"
            ):
                run_workflow(
                    git_url,
                    git_branch,
                    ssh_key_path=ssh_key_path,
                    docker_image_override=image
                )
            console.rule(characters="=", style="green")
        except BodyworkWorkflowExecutionError:
            sys.exit(1)
    else:
        if not asynchronous_job_name:
            async_deployment_job_name = make_valid_k8s_name(
                f"async-workflow-{git_url}.{git_branch}."
                f"{datetime.now().isoformat(timespec='seconds')}"
            )
        else:
            async_deployment_job_name = f"async-workflow-{asynchronous_job_name}"
        print_info("Using asynchronous workflow controller.")
        create_workflow_job(
            BODYWORK_NAMESPACE,
            async_deployment_job_name,
            git_url,
            git_branch,
            retries,
            image if image else BODYWORK_DOCKER_IMAGE,
            ssh_key_path,
            secrets_group,
        )
        sys.exit(0)


@get.command("deployment")
@get.command("deployments")
@handle_k8s_exceptions
@k8s_auth
def _get_deployment(
    name: str = Argument(None),
    service_name: Optional[str] = Argument(None),
    asynchronous: bool = Option(False, "--async"),
    logs: str = Option(""),
    namespace: Optional[str] = Option(None),
):
    if asynchronous:
        if logs:
            display_workflow_job_logs(BODYWORK_NAMESPACE, logs)
        else:
            display_workflow_job_history(
                BODYWORK_NAMESPACE, "async-workflow"
            )
    else:
        display_deployments(namespace, name, service_name)
    sys.exit(0)


@update.command("deployment")
@handle_k8s_exceptions
@k8s_auth
def _update_deployment(
    git_url: str = Argument(...),
    git_branch: str = Argument(...),
    asynchronous: bool = Option(False, "--async"),
    asynchronous_job_name: str = Option("", "--async-job-name"),
    image: Optional[str] = Option(None, "--bodywork-image"),
    retries: int = Option(1),
):
    _create_deployment(
        git_url=git_url,
        git_branch=git_branch,
        asynchronous=asynchronous,
        asynchronous_job_name=asynchronous_job_name,
        ssh_key_path="",
        secrets_group="",
        image=image,
        retries=retries,
    )


@delete.command("deployment")
@handle_k8s_exceptions
@k8s_auth
def _delete_deployment(
    name: str = Argument(...), asynchronous: bool = Option(False, "--async")
):
    if asynchronous:
        delete_workflow_job(BODYWORK_NAMESPACE, name)
    else:
        delete_deployment(name)
    sys.exit(0)


@create.command("cronjob")
@handle_k8s_exceptions
@k8s_auth
def _create_cronjob(
    git_url: str = Argument(...),
    git_branch: str = Argument(...),
    schedule: str = Option(...),
    name: str = Option(...),
    retries: int = Option(1),
    history_limit: int = Option(1),
    ssh_key_path: str = Option("", "--ssh"),
    secrets_group: str = Option("", "--group", "--secrets-group"),
):
    create_workflow_cronjob(
        BODYWORK_NAMESPACE,
        schedule,
        make_valid_k8s_name(name),
        git_url,
        git_branch if git_branch else "master",
        retries,
        history_limit,
        ssh_key_path,
        secrets_group,
    )
    sys.exit(0)


@get.command("cronjob")
@get.command("cronjobs")
@handle_k8s_exceptions
@k8s_auth
def _get_cronjob(
    name: Optional[str] = Argument(None),
    history: bool = Option(False),
    logs: str = Option(""),
):
    if name and history and not logs:
        display_workflow_job_history(BODYWORK_NAMESPACE, name)
    elif name and not history and logs:
        display_workflow_job_logs(BODYWORK_NAMESPACE, logs)
    elif name and history and logs:
        print_warn("Cannot specify both --logs and --history.")
        sys.exit(1)
    else:
        display_cronjobs(BODYWORK_NAMESPACE, name)
    sys.exit(0)


@update.command("cronjob")
@handle_k8s_exceptions
@k8s_auth
def _update_cronjob(
    git_url: str = Argument(...),
    git_branch: str = Argument(...),
    schedule: str = Option(...),
    name: str = Option(...),
    retries: int = Option(1),
    history_limit: int = Option(1),
):
    update_workflow_cronjob(
        BODYWORK_NAMESPACE,
        name,
        schedule,
        git_url,
        git_branch,
        retries,
        history_limit,
    )
    sys.exit(0)


@delete.command("cronjob")
@handle_k8s_exceptions
@k8s_auth
def _delete_cronjob(name: str = Argument(...)):
    delete_workflow_cronjob(BODYWORK_NAMESPACE, name)
    sys.exit(0)


@create.command("secret")
@handle_k8s_exceptions
@k8s_auth
def _create_secret(
    name: str = Argument(...), group: str = Option(...), data: List[str] = Option(...)
):
    try:
        var_names_and_values = parse_cli_secrets_strings(data)
        create_secret(
            BODYWORK_NAMESPACE, group, name, var_names_and_values
        )
    except ValueError:
        print_warn(
            "Could not parse secret data - example format: --data USERNAME=alex "
            "PASSWORD=alex123"
        )
        sys.exit(1)
    sys.exit(0)


@get.command("secret")
@get.command("secrets")
@handle_k8s_exceptions
@k8s_auth
def _get_secret(
    name: Optional[str] = Argument(None), group: Optional[str] = Option(None)
):
    if name and not group:
        print_warn("Please specify which secrets group the secret belongs to.")
        sys.exit(1)
    else:
        display_secrets(
            BODYWORK_NAMESPACE,
            group,
            name,
        )
        sys.exit(0)


@update.command("secret")
@handle_k8s_exceptions
@k8s_auth
def _update_secret(
    name: str = Argument(...), group: str = Option(...), data: List[str] = Option(...)
):
    try:
        var_names_and_values = parse_cli_secrets_strings(data)
        update_secret(
            BODYWORK_NAMESPACE, group, name, var_names_and_values
        )
        sys.exit(0)
    except ValueError:
        print_warn(
            "Could not parse secret data - example format: --data USERNAME=alex "
            "PASSWORD=alex123"
        )
        sys.exit(1)


@delete.command("secret")
@handle_k8s_exceptions
@k8s_auth
def _delete_secret(name: str = Argument(...), group: str = Option(...)):
    if name and not group:
        print_warn("Please specify which secrets group the secret belongs to.")
        sys.exit(1)
    else:
        delete_secret(BODYWORK_NAMESPACE, group, name)
        sys.exit(0)



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
This module contains functions for helping with the creation of a
Kubernetes namespace with the necessary service accounts and roles to
run workflows, jobs and service deployments, securely. It is target at
being called from the CLI.
"""
from .terminal import print_info, print_warn
from .. import k8s
from ..constants import (
    BODYWORK_WORKFLOW_SERVICE_ACCOUNT,
)


def is_namespace_available_for_bodywork(namespace: str) -> bool:
    """Is the namespace available for use by Bodywork.

    :param namespace: The namespace to check.
    :return: Boolean flag indicating the namespace's availability for
        running Bodywork projects.
    """
    if not k8s.namespace_exists(namespace):
        print_warn(f"Could not find namespace={namespace} on k8s cluster.")
        return False
    workflow_controller_sa_exists = k8s.service_account_exists(
        namespace, BODYWORK_WORKFLOW_SERVICE_ACCOUNT
    )
    workflow_controller_sa_cluster_role_binding_exists = (
        k8s.cluster_role_binding_exists(
            k8s.workflow_cluster_role_binding_name(namespace)
        )
    )
    is_namespace_setup = (
        True
        if (
            workflow_controller_sa_exists
            and workflow_controller_sa_cluster_role_binding_exists
        )
        else False
    )
    if is_namespace_setup:
        return True
    else:
        if not workflow_controller_sa_exists:
            print_warn(
                f"Missing service-account={BODYWORK_WORKFLOW_SERVICE_ACCOUNT} from "
                f"namespace={namespace}."
            )
        if not workflow_controller_sa_cluster_role_binding_exists:
            print_warn(
                f"Missing cluster-role-binding="
                f"{k8s.workflow_cluster_role_binding_name(namespace)}."
            )
        return False


def setup_namespace_with_service_accounts_and_roles(namespace: str) -> None:
    """Setup kubernetes namespace for Bodywork Workflow accounts.

    If the namespace does not already exist, then it will be created
    first. Then the cluster/service accounts required by bodywork to
    run workflows will be created.

    Note, that to use this function the Kubernetes user running the
    command must be authorised to create namespaces, service accounts,
    roles and cluster-roles.

    :param namespace: Name of namespace.
    """
    if k8s.namespace_exists(namespace):
        print_warn(f"namespace={namespace} already exists.")
    else:
        print_info(f"Creating namespace={namespace}.")
        k8s.create_namespace(namespace)

    workflow_sa = BODYWORK_WORKFLOW_SERVICE_ACCOUNT
    workflow_crb = k8s.workflow_cluster_role_binding_name(namespace)
    if k8s.service_account_exists(namespace, workflow_sa):
        print_warn(
            f"service-account={workflow_sa} already exists in namespace={namespace}."
        )
    else:
        print_info(f"Creating service-account={workflow_sa} in namespace={namespace}.")
        print_info(f"Creating cluster-role-binding={workflow_crb}.")
        k8s.setup_workflow_service_accounts(namespace)



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
This module contains functions for managing Kubernetes workflow jobs.
They are targeted for use via the CLI.
"""
import re
from typing import Optional
from pathlib import Path

from .terminal import print_dict, print_info, print_pod_logs, print_warn
from .. import k8s
from ..constants import BODYWORK_DOCKER_IMAGE, SSH_SECRET_NAME


def create_workflow_job(
    namespace: str,
    job_name: str,
    project_repo_url: str,
    project_repo_branch: str = "master",
    retries: int = 2,
    image: str = BODYWORK_DOCKER_IMAGE,
    ssh_key_path: str = None,
    secrets_group: str = None,
) -> None:
    """Create a new workflow job within a namespace.

    :param namespace: The namespace to deploy the job to.
    :param job_name: The name of the Bodywork job.
    :param project_repo_url: The URL for the Bodywork project Git
        repository.
    :param project_repo_branch: The branch of the Bodywork project Git
        repository that will be used as the executable codebase,
        defaults to 'master'.
    :param retries: Number of times to retry running the stage to
        completion (if necessary), defaults to 2.
    :param image: Docker image to use for running the stages within,
        defaults to BODYWORK_DOCKER_IMAGE.
    :param ssh_key_path: SSH key filepath.
    :param secrets_group: Secrets group to use if using SSH.
    """
    if not k8s.namespace_exists(namespace):
        print_warn(f"Could not find namespace={namespace} on k8s cluster.")
        return None
    if _is_existing_workflow_job(namespace, job_name):
        print_warn(f"Cannot create workflow-job={job_name} as it already exists.")
        return None
    if ssh_key_path:
        if not secrets_group:
            print_warn("Please specify Secrets Group in config to use SSH.")
            return None
        try:
            k8s.create_ssh_key_secret_from_file(secrets_group, Path(ssh_key_path))
        except FileNotFoundError:
            print_warn(f"Could not find SSH key file at: {ssh_key_path}")
            return None
    if secrets_group:
        if not k8s.secret_exists(
            namespace, k8s.create_complete_secret_name(SSH_SECRET_NAME, secrets_group)
        ):
            print_warn(
                f"Could not find SSH secret: {SSH_SECRET_NAME} in group: {secrets_group}"
            )
        env_vars = [k8s.create_secret_env_variable(secrets_group)]
    else:
        env_vars = None

    configured_job = k8s.configure_workflow_job(
        namespace,
        project_repo_url,
        project_repo_branch,
        retries,
        image,
        job_name,
        container_env_vars=env_vars,
    )
    k8s.create_workflow_job(configured_job)
    print_info(f"Created workflow-job={job_name}.")


def delete_workflow_job(namespace: str, job_name: str) -> None:
    """Delete workflow job from a specific namespace.

    :param namespace: Namespace where the job resides.
    :param job_name: Name of the job to delete.
    """
    if not k8s.namespace_exists(namespace):
        print_warn(f"Could not find namespace={namespace} on k8s cluster.")
        return None
    if not _is_existing_workflow_job(namespace, job_name):
        print_warn(f"Could not find workflow-job={job_name}.")
        return None
    k8s.delete_job(namespace, job_name)
    print_info(f"Deleted workflow-job={job_name}.")


def create_workflow_cronjob(
    namespace: str,
    schedule: str,
    job_name: str,
    project_repo_url: str,
    project_repo_branch: str = "master",
    retries: int = 2,
    workflow_job_history_limit: int = 1,
    ssh_key_path: str = None,
    secrets_group: str = None,
) -> None:
    """Create a new cronjob within a namespace.

    :param namespace: The namespace to deploy the job to.
    :param schedule: A valid cron schedule definition.
    :param job_name: The name of the Bodywork the job to create.
    :param project_repo_url: The URL for the Bodywork project Git
        repository.
    :param project_repo_branch: The branch of the Bodywork project Git
        repository that will be used as the executable codebase,
        defaults to 'master'.
    :param retries: Number of times to retry running the stage to
        completion (if necessary), defaults to 2.
    :param workflow_job_history_limit: Minimum number of
        historical workflow jobs, so logs can be retrieved.
    :param ssh_key_path: SSH key filepath.
    :param secrets_group: Secrets group to use if using SSH.
    """
    if not k8s.namespace_exists(namespace):
        print_warn(f"Could not find namespace={namespace} on k8s cluster.")
        return None
    if _is_existing_workflow_cronjob(namespace, job_name):
        print_warn(f"Cannot create cronjob={job_name} as it already exists.")
        return None
    if not _is_valid_cron_schedule(schedule):
        print_warn(f"Invalid cronjob schedule: {schedule}.")
        return None
    if ssh_key_path:
        if not secrets_group:
            print_warn("Please specify Secrets Group in config to use SSH.")
            return None
        try:
            k8s.create_ssh_key_secret_from_file(secrets_group, Path(ssh_key_path))
        except FileNotFoundError:
            print_warn(f"Could not find SSH key file at: {ssh_key_path}")
            return None
    if secrets_group:
        if not k8s.secret_exists(
            namespace, k8s.create_complete_secret_name(SSH_SECRET_NAME, secrets_group)
        ):
            print_warn(
                f"Could not find SSH secret: {SSH_SECRET_NAME} in group: {secrets_group}"
            )
        env_vars = [k8s.create_secret_env_variable(secrets_group)]
    else:
        env_vars = None
    configured_job = k8s.configure_workflow_cronjob(
        schedule,
        namespace,
        job_name,
        project_repo_url,
        project_repo_branch,
        retries,
        workflow_job_history_limit,
        workflow_job_history_limit,
        env_vars=env_vars,
    )
    k8s.create_workflow_cronjob(configured_job)
    print_info(f"Created cronjob={job_name}.")


def update_workflow_cronjob(
    namespace: str,
    job_name: str,
    schedule: Optional[str] = None,
    project_repo_url: Optional[str] = None,
    project_repo_branch: Optional[str] = None,
    retries: Optional[int] = None,
    workflow_job_history_limit: Optional[int] = None,
) -> None:
    """Update a new cronjob within a namespace.

    :param namespace: The namespace the job resides in.
    :param job_name: The name of the cronjob.
    :param schedule: A valid cron schedule definition.
    :param project_repo_url: The URL for the Bodywork project Git
        repository.
    :param project_repo_branch: The branch of the Bodywork project Git
        repository that will be used as the executable codebase,
        defaults to 'master'.
    :param retries: Number of times to retry running the stage to
        completion (if necessary), defaults to 2.
    :param workflow_job_history_limit: Minimum number of
        historical workflow jobs, so logs can be retrieved.
    """
    if not k8s.namespace_exists(namespace):
        print_warn(f"Could not find namespace={namespace} on k8s cluster.")
        return None
    if not _is_existing_workflow_cronjob(namespace, job_name):
        print_warn(f"Could not find cronjob={job_name}.")
        return None
    if schedule and not _is_valid_cron_schedule(schedule):
        print_warn(f"Invalid cronjob schedule: {schedule}.")
        return None
    k8s.update_workflow_cronjob(
        namespace,
        job_name,
        schedule,
        project_repo_url,
        project_repo_branch,
        retries,
        workflow_job_history_limit,
        workflow_job_history_limit,
    )
    print_info(f"Updated cronjob={job_name}.")


def delete_workflow_cronjob(namespace: str, job_name: str) -> None:
    """Create a new cronjob within a k8s namespace.

    :param namespace: The namespace where the cronjob resides.
    :param job_name: The name of the cronjob to be deleted.
    """
    if not k8s.namespace_exists(namespace):
        print_warn(f"Could not find namespace={namespace} on k8s cluster.")
        return None
    if not _is_existing_workflow_cronjob(namespace, job_name):
        print_warn(f"Could not find cronjob={job_name}.")
        return None
    k8s.delete_workflow_cronjob(namespace, job_name)
    print_info(f"Deleted cronjob={job_name}.")


def display_cronjobs(namespace: str, job_name: Optional[str] = None) -> None:
    """Print cronjobs to stdout.

    :param namespace: Namespace in which to look for cronjobs.
    :param job_name: Name of cronjob resource, defaults to None.
    """
    if not k8s.namespace_exists(namespace):
        print_warn(f"Could not find namespace={namespace} on k8s cluster.")
        return None
    cronjobs_info = k8s.list_workflow_cronjobs(namespace)
    if job_name:
        print_dict(cronjobs_info[job_name], job_name)
    else:
        table_data = {name: data["git_url"] for name, data in cronjobs_info.items()}
        print_dict(table_data, "cronjobs", "Name", "Git Repository URL")


def display_workflow_job_history(namespace: str, job_name: str) -> None:
    """Print info on workflow jobs, triggered by a cronjob, to stdout.

    :param namespace: Namespace in which to look for cronjobs.
    :param job_name: Name of the cronjob.
    """
    if not k8s.namespace_exists(namespace):
        print_warn(f"Could not find namespace={namespace} on k8s cluster.")
        return None
    workflow_jobs_info = k8s.list_workflow_jobs(namespace, job_name)
    for name, data in workflow_jobs_info.items():
        print_dict(data, f"workflow job = {name}")


def display_workflow_job_logs(namespace: str, job_name: str) -> None:
    """Print workflow job logs to stdout.

    :param namespace: Namespace in which the workflow job exists.
    :param job_name: The full name of the specific workflow job
        executed - e.g. NAME_OF_PROJECT-12345.
    """
    if not k8s.namespace_exists(namespace):
        print_warn(f"Could not find namespace={namespace} on k8s cluster.")
        return None
    workflow_job_pod_name = k8s.get_latest_pod_name(namespace, job_name)
    if workflow_job_pod_name is None:
        print_warn(f"Cannot find pod for workflow job={job_name}.")
        return None
    workflow_job_logs = k8s.get_pod_logs(namespace, workflow_job_pod_name)
    print_pod_logs(workflow_job_logs, f"logs for workflow execution = {job_name}")


def _is_existing_workflow_job(namespace: str, job_name: str) -> bool:
    """Can the named job be found in the namespace.

    :param namespace: The namespace to look in.
    :param job_name: The name of the Bodywork the job.
    :return: A boolean flag.
    """
    jobs_in_namespace = k8s.list_workflow_jobs(namespace, job_name)
    return True if job_name in jobs_in_namespace.keys() else False


def _is_existing_workflow_cronjob(namespace: str, job_name: str) -> bool:
    """Can the named cronjob be found in the namespace.

    :param namespace: The namespace to look in.
    :param job_name: The name of the Bodywork the job.
    :return: A boolean flag.
    """
    cronjobs_in_namespace = k8s.list_workflow_cronjobs(namespace)
    return True if job_name in cronjobs_in_namespace.keys() else False


def _is_valid_cron_schedule(schedule: str) -> bool:
    """Does a string represent a valid cron schedule.

    :param schedule: A string describing a cron schedule.
    :return: A boolean flag.
    """
    parsed_schedule = [e for e in schedule.split(" ")]
    if len(parsed_schedule) != 5:
        return False

    minutes_pattern_matches = re.fullmatch(
        r"^([1-5]?[0-9](,|$))+" r"|^(\*|[1-5]?[0-9]-[1-5]?[0-9])(\/[1-5]?[0-9]$|$)",
        parsed_schedule[0],
    )
    if minutes_pattern_matches is None:
        return False

    hours_pattern_matches = re.fullmatch(
        r"^((2[0-3]|1?[0-9])(,|$))+"
        r"|^(\*|(2[0-3]|1?[0-9])-(2[0-3]|1?[0-9]))(\/(2[0-3]|1?[0-9])$|$)",
        parsed_schedule[1],
    )
    if hours_pattern_matches is None:
        return False

    day_of_the_month_pattern_matches = re.fullmatch(
        r"^((3[0-1]|[1-2]?[0-9])(,|$))+"
        r"|^(\*|(3[0-1]|[1-2]?[0-9])-(3[0-1]|[1-2]?[0-9]))(\/(3[0-1]|[1-2]?[0-9])$|$)",  # noqa
        parsed_schedule[2],
    )
    if day_of_the_month_pattern_matches is None:
        return False

    month_pattern_matches = re.fullmatch(
        r"^((1[0-2]|[0-9])(,|$))+"
        r"|^(\*|(1[0-2]|[0-9])-(1[0-2]|[0-9]))(\/(1[0-2]|[0-9])$|$)",
        parsed_schedule[3],
    )
    if month_pattern_matches is None:
        return False

    day_of_week_pattern_matches = re.fullmatch(
        r"^([0-6](,|$))+" r"|^(\*|[0-6]-[0-6])(\/[0-6]$|$)", parsed_schedule[4]
    )
    if day_of_week_pattern_matches is None:
        return False

    return True



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Helper functions for working with the Kubernetes API.
"""
import json
import re
from typing import cast, List, Tuple

from kubernetes.client.rest import ApiException
from kubernetes import client as k8s


EnvVars = k8s.V1EnvVar


def api_exception_msg(e: ApiException) -> str:
    """Get k8s API error message from exception object.

    :param e: Kubernetes API exception
    :return: Error message returned by the k8s API.
    """
    try:
        body = json.loads(e.body)
        message = body["message"]
        return cast(str, message)
    except (KeyError, TypeError):
        return ""


def make_valid_k8s_name(name: str) -> str:
    """Remove invalid characters from k8s resource name.

    :param name: Original intended name.
    :return: Valid Kubernetes resource name.
    """
    return re.sub(r"[^a-zA-Z0-9.]+", "-", name.strip())


def create_k8s_environment_variables(
    key_value_pairs: List[Tuple[str, str]]
) -> List[k8s.V1EnvVar]:
    """Creates K8s environment variable from key/value pairs.

    :param key_value_pairs: Environment variables to create.
    :return: List of K8s environment variables.
    """
    return [k8s.V1EnvVar(name=name, value=value) for name, value in key_value_pairs]



from typing import Any, Dict, Optional

from rich.console import Console
from rich.table import Table

console = Console(highlight=False, soft_wrap=False, width=175)


def print_info(msg: str) -> None:
    """Print an info message."""
    console.print(msg, style="green")


def print_warn(msg: str) -> None:
    """Print a warning message."""
    console.print(msg, style="red")


def print_dict(
    the_dict: Dict[str, Any],
    table_name: Optional[str] = None,
    key_col_name: str = "Field",
    val_col_name: str = "Value",
) -> None:
    """Render dict as a table in terminal.

    :param the_dict: The dictionary to render.
    :param table_name: Table name, defaults to None.
    :param key_col_name: Header for the keys column, defaults to 'Field'.
    :param val_col_name: Header for the values column, defaults to 'Value'.
    """
    table = Table(title=f"{table_name if table_name else ''}", title_style="bold")
    table.add_column(f"[yellow]{key_col_name}[/yellow]", style="bold purple")
    table.add_column(f"[yellow]{val_col_name}[/yellow]", style="bold green")
    for field, value in the_dict.items():
        table.add_row(str(field), str(value))
    console.print(table)


def print_pod_logs(logs: str, header: str) -> None:
    """Render pod logs.

    :param logs: The logs!
    :param header: Text to associate with the logs.
    """
    console.rule(f"[yellow]{header}[/yellow]", style="yellow")
    console.print(logs, style="grey58")
    console.rule(style="yellow")



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
High-level interface to the Kubernetes APIs used to manage
authentication and authorisation for cluster resources.
"""
import os

from kubernetes import client as k8s, config as k8s_config

from ..constants import (
    BODYWORK_WORKFLOW_CLUSTER_ROLE,
    BODYWORK_WORKFLOW_SERVICE_ACCOUNT,
    BODYWORK_STAGES_SERVICE_ACCOUNT,
)


def load_kubernetes_config() -> None:
    """Attempt to load k8s config from file.

    If running within a k8s cluster, then KUBERNETES_SERVICE_HOST will
    be present, which allows us to call a special initialization method
    for in-cluster situations. Otherwise the standard ~/.kube/config is
    read.

    :raises RuntimeError: if a kubeconfig cannot be loaded.
    """
    if os.getenv("KUBERNETES_SERVICE_HOST"):
        k8s_config.load_incluster_config()
    else:
        k8s_config.load_kube_config()


def service_account_exists(namespace: str, name: str) -> bool:
    """Does the service-account exist within the namespace.

    :param namespace: Kubernetes namespace to check.
    :param name: The name of the service-account to check.
    :return: True if the service-account was found, otherwise False.
    """
    service_account_objects = (
        k8s.CoreV1Api().list_namespaced_service_account(namespace=namespace).items
    )
    service_account_names = [
        service_account_object.metadata.name
        for service_account_object in service_account_objects
    ]
    return True if name in service_account_names else False


def cluster_role_exists(name: str) -> bool:
    """Does the cluster-role exist.

    :param name: The name of the cluster-role to check.
    :return: True if the cluster-role was found, otherwise False.
    """
    cluster_role_objects = k8s.RbacAuthorizationV1Api().list_cluster_role().items
    cluster_role_names = [
        cluster_role_object.metadata.name
        for cluster_role_object in cluster_role_objects
    ]
    return True if name in cluster_role_names else False


def workflow_cluster_role_binding_name(namespace: str) -> str:
    """Get cluster-role-binding name for workflow service-account.

    :param namespace: The namespace in which a workflow service-account
        has been deployed to.
    :return: The name assigned to the cluster-role-binding that
        associates the workflow cluster-role to the workflow
        service-account, within a namespace setup for Bodywork.
    """
    return f"{BODYWORK_WORKFLOW_CLUSTER_ROLE}--{namespace}"


def cluster_role_binding_exists(name: str) -> bool:
    """Does the cluster-role-binding exist.

    :param name: The name of the cluster-role-binding to check.
    :return: True if the cluster-role-binding was found, otherwise False.
    """
    cluster_role_binding_objects = (
        k8s.RbacAuthorizationV1Api().list_cluster_role_binding().items
    )
    cluster_role_binding_names = [
        cluster_role_binding_object.metadata.name
        for cluster_role_binding_object in cluster_role_binding_objects
    ]
    return True if name in cluster_role_binding_names else False


def delete_cluster_role_binding(name: str) -> None:
    """Delete a cluster-role-binding.

    :param name: The name assigned to the cluster-role-binding.
    """
    k8s.RbacAuthorizationV1Api().delete_cluster_role_binding(name=name)


def setup_workflow_service_accounts(namespace: str) -> None:
    """Setup a workflow controller service-account with required roles.

    :param namespace: Namespace in which the service-account will be
        placed.
    """
    service_account_object = k8s.V1ServiceAccount(
        metadata=k8s.V1ObjectMeta(
            namespace=namespace, name=BODYWORK_WORKFLOW_SERVICE_ACCOUNT
        )
    )
    k8s.CoreV1Api().create_namespaced_service_account(
        namespace=namespace, body=service_account_object
    )

    if not cluster_role_exists(BODYWORK_WORKFLOW_CLUSTER_ROLE):
        cluster_role_object = k8s.V1ClusterRole(
            metadata=k8s.V1ObjectMeta(name=BODYWORK_WORKFLOW_CLUSTER_ROLE),
            rules=[
                k8s.V1PolicyRule(
                    api_groups=[""],
                    resources=["namespaces", "services"],
                    verbs=["get", "list", "create", "delete"],
                ),
                k8s.V1PolicyRule(
                    api_groups=["rbac.authorization.k8s.io"],
                    resources=["clusterrolebindings"],
                    verbs=["get", "list", "create"],
                ),
                k8s.V1PolicyRule(
                    api_groups=[""],
                    resources=["serviceaccounts"],
                    verbs=["list", "create"],
                ),
                k8s.V1PolicyRule(
                    api_groups=["rbac.authorization.k8s.io"],
                    resources=["roles", "rolebindings"],
                    verbs=["create"],
                ),
                k8s.V1PolicyRule(
                    api_groups=[""],
                    resources=["configmaps"],
                    verbs=["get", "list"],
                ),
                k8s.V1PolicyRule(
                    api_groups=[""],
                    resources=["secrets"],
                    verbs=["get", "list", "create", "update", "patch"],
                ),
                k8s.V1PolicyRule(
                    api_groups=["apps", "batch"], resources=["*"], verbs=["*"]
                ),
                k8s.V1PolicyRule(
                    api_groups=["networking.k8s.io", "extensions"],
                    resources=["ingresses"],
                    verbs=["*"],
                ),
            ],
        )
        k8s.RbacAuthorizationV1Api().create_cluster_role(body=cluster_role_object)

    if not cluster_role_binding_exists(workflow_cluster_role_binding_name(namespace)):
        cluster_role_binding_object = k8s.V1ClusterRoleBinding(
            metadata=k8s.V1ObjectMeta(
                name=workflow_cluster_role_binding_name(namespace), namespace=namespace
            ),
            role_ref=k8s.V1RoleRef(
                kind="ClusterRole",
                name=BODYWORK_WORKFLOW_CLUSTER_ROLE,
                api_group="rbac.authorization.k8s.io",
            ),
            subjects=[
                k8s.V1Subject(
                    kind="ServiceAccount",
                    name=BODYWORK_WORKFLOW_SERVICE_ACCOUNT,
                    namespace=namespace,
                )
            ],
        )
        k8s.RbacAuthorizationV1Api().create_cluster_role_binding(
            body=cluster_role_binding_object
        )


def setup_stages_service_account(namespace: str) -> None:
    """Setup a service-account with required roles for jobs-and-deployments.

    :param namespace: Namespace in which the service-account will be
        placed.
    """
    service_account_object = k8s.V1ServiceAccount(
        metadata=k8s.V1ObjectMeta(
            namespace=namespace, name=BODYWORK_STAGES_SERVICE_ACCOUNT
        )
    )
    k8s.CoreV1Api().create_namespaced_service_account(
        namespace=namespace, body=service_account_object
    )

    role_object = k8s.V1Role(
        metadata=k8s.V1ObjectMeta(
            namespace=namespace, name=BODYWORK_STAGES_SERVICE_ACCOUNT
        ),
        rules=[
            k8s.V1PolicyRule(
                api_groups=[""],
                resources=["secrets", "configmaps"],
                verbs=["get", "list"],
            )
        ],
    )
    k8s.RbacAuthorizationV1Api().create_namespaced_role(
        namespace=namespace, body=role_object
    )

    role_binding_object = k8s.V1RoleBinding(
        metadata=k8s.V1ObjectMeta(
            namespace=namespace, name=BODYWORK_STAGES_SERVICE_ACCOUNT
        ),
        role_ref=k8s.V1RoleRef(
            kind="Role",
            name=BODYWORK_STAGES_SERVICE_ACCOUNT,
            api_group="rbac.authorization.k8s.io",
        ),
        subjects=[
            k8s.V1Subject(
                kind="ServiceAccount",
                name=BODYWORK_STAGES_SERVICE_ACCOUNT,
                namespace=namespace,
            )
        ],
    )
    k8s.RbacAuthorizationV1Api().create_namespaced_role_binding(
        namespace=namespace, body=role_binding_object
    )



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
This module contains functions for managing service deployments that have
been created as part of executed workflows.
"""
from typing import Optional

from .terminal import print_dict, print_info, print_warn
from .. import k8s


def display_deployments(
    namespace: Optional[str] = None,
    name: Optional[str] = None,
    service_name: Optional[str] = None,
) -> None:
    """Print active deployments to stdout.

    :param namespace: Namespace in which to look for deployments.
    :param name: Name of the deployment to display.
    :param service_name: Name of service to display.
    """
    if namespace and not k8s.namespace_exists(namespace):
        print_warn(f"Could not find namespace={namespace} on k8s cluster.")
        return None
    deployments = k8s.list_service_stage_deployments(namespace, name)
    if not deployments:
        print("No deployments found")
        return None
    if service_name and name:
        id = k8s.deployment_id(name, service_name)
        if id not in deployments:
            print_warn(f"Could not find service={service_name}.")
            return None
        print_dict(
            deployments[id],
            id,
        )
    else:
        table_data = {svc_key: data["git_url"] for svc_key, data in deployments.items()}
        print_dict(
            table_data,
            "operational services",
            "Deployment Name / Service Name",
            "Git Repository URL",
        )


def delete_service_deployment_in_namespace(namespace: str, name: str) -> None:
    """Delete a service deployment from within a k8s namespace.

    :param namespace: The namespace in which the service deployment
        exists.
    :param name: The name of the service deployment to delete.
    """
    if not k8s.namespace_exists(namespace):
        print_warn(f"Could not find namespace={namespace} on k8s cluster.")
        return None
    if name not in k8s.list_service_stage_deployments(namespace).keys():
        print_warn(f"Could not find service={name}.")
        return None
    k8s.delete_deployment(namespace, name)
    print_info(f"Deleted service={name}.")
    if k8s.is_exposed_as_cluster_service(namespace, name):
        k8s.stop_exposing_cluster_service(namespace, name)
        print_info(
            f"Stopped exposing service at {k8s.cluster_service_url(namespace, name)}"
        )
    if k8s.has_ingress(namespace, name):
        k8s.delete_deployment_ingress(namespace, name)
        print_info(
            f"Deleted ingress to service at {k8s.ingress_route(namespace, name)}"
        )


def delete_deployment(deployment_name: str) -> None:
    """Delete a deployment by deleting the namespace it's in.

    :param deployment_name: The name of the deployment.
    """
    deployments = k8s.list_service_stage_deployments(name=deployment_name)
    if not deployments:
        print_info(f"deployment={deployment_name} could not be found on k8s cluster.")
        return None
    k8s.delete_namespace(list(deployments.values())[0]["namespace"])
    print_info(f"deployment={deployment_name} deleted.")



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
This module contains functions for downloading Bodywork projects from
Git repositories.
"""
import os
import re
from enum import Enum
from pathlib import Path
from subprocess import run, CalledProcessError, DEVNULL, PIPE
from urllib.parse import urlparse

from .exceptions import BodyworkGitError
from .constants import (
    DEFAULT_PROJECT_DIR,
    SSH_DIR_NAME,
    SSH_PRIVATE_KEY_ENV_VAR,
    GITHUB_SSH_FINGERPRINT,
    GITLAB_SSH_FINGERPRINT,
    BITBUCKET_SSH_FINGERPRINT,
    AZURE_SSH_FINGERPRINT,
    GIT_SSH_COMMAND,
    DEFAULT_SSH_FILE,
)
from .logs import bodywork_log_factory

_log = bodywork_log_factory()


def download_project_code_from_repo(
    url: str,
    branch: str = "master",
    destination: Path = DEFAULT_PROJECT_DIR,
    ssh_key_path: str = None,
) -> None:
    """Download Bodywork project code from Git repository,

    :param url: Git repository URL.
    :param branch: The Git branch to download, defaults to 'master'.
    :param destination: The name of the directory int which the
        repository will be cloned, defaults to DEFAULT_PROJECT_DIR.
    :param ssh_key_path: SSH key filepath.
    :raises BodyworkGitError: If Git is not available on the system or the
        Git repository cannot be accessed.
    """
    try:
        run(["git", "--version"], check=True, stdout=DEVNULL)
    except CalledProcessError:
        raise BodyworkGitError("git is not available")
    try:
        if get_connection_protocol(url) is ConnectionProtocol.SSH:
            hostname = urlparse(f"ssh://{url}").hostname
            if hostname:
                setup_ssh_for_git_host(hostname, ssh_key_path)
            else:
                raise ValueError(
                    f"Unable to derive hostname from URL {url}. Please check "
                )
    except Exception as e:
        msg = f"Unable to setup SSH for Git and you are trying to connect via SSH: {e}"
        raise BodyworkGitError(msg)
    try:
        run(
            ["git", "clone", "--branch", branch, "--single-branch", url, destination],
            check=True,
            encoding="utf-8",
            stdout=DEVNULL,
            stderr=PIPE,
        )
    except CalledProcessError as e:
        msg = f"Git clone failed - calling {e.cmd} returned {e.stderr}"
        raise BodyworkGitError(msg)


class ConnectionProtocol(Enum):
    """Connection protocol used to access Git repo."""

    FILE = "file"
    HTTPS = "https"
    SSH = "ssh"


def get_connection_protocol(connection_string: str) -> ConnectionProtocol:
    """Derive connection protocol used to retrieve Git repo.

    :param connection_string: The string containing the connection
        details for the remote Git repository - e.g. the GitHub URL.
    :raises RuntimeError: if the connection protocol cannot be
        identified or is not supported.
    :return: The connection protocol type.
    """
    if re.match("^https://", connection_string):
        return ConnectionProtocol.HTTPS
    elif re.match("^git@", connection_string):
        return ConnectionProtocol.SSH
    elif re.match("^file://", connection_string):
        return ConnectionProtocol.FILE
    else:
        msg = (
            f"cannot identify connection protocol in {connection_string}"
            f"- currently, there is only support for HTTPS and SSH"
        )
        raise RuntimeError(msg)


def setup_ssh_for_git_host(hostname: str, ssh_key_path: str = None) -> None:
    """Setup system for SSH interaction with GitHub.

    Using the private key assigned to an environment variable, this
    function creates a new SSH configuration in the working directory
    and then tells Git to use it for SSH by exporting the
    GIT_SSH_COMMAND environment variable.

    :param hostname: Hostname to SSH to.
    :param ssh_key_path: SSH key file to use.
    """
    ssh_dir = Path.home() / SSH_DIR_NAME
    if SSH_PRIVATE_KEY_ENV_VAR in os.environ:
        _log.info("Using SSH key from environment variable.")
        try:
            private_key = ssh_dir / DEFAULT_SSH_FILE
            ssh_dir.mkdir(mode=0o700, exist_ok=True)
            private_key.touch(0o700, exist_ok=True)
            key = os.environ[SSH_PRIVATE_KEY_ENV_VAR]
            if key[-1] != "\n":
                key = f"{key}\n"
            with Path(private_key).open(mode="w", newline="\n") as file_handle:
                file_handle.write(key)
        except OSError as e:
            raise RuntimeError(
                f"Unable to create private key {private_key} from"
                f" {SSH_PRIVATE_KEY_ENV_VAR} environment variable."
            ) from e
    elif ssh_key_path:
        private_key = Path(ssh_key_path)
        if not private_key.exists():
            msg = f"Failed to setup SSH for {hostname} - cannot find SSH key {ssh_key_path}"  # noqa
            raise FileNotFoundError(msg)
    else:
        msg = f"Failed to setup SSH for {hostname} - cannot find SSH keys or {SSH_PRIVATE_KEY_ENV_VAR} environment variable."  # noqa
        raise RuntimeError(msg)

    _configure_known_hosts(hostname, ssh_dir)
    os.environ[GIT_SSH_COMMAND] = f"ssh -i '{private_key}'" f" -o IdentitiesOnly=yes"


def _configure_known_hosts(hostname, ssh_dir):
    try:
        known_hosts = ssh_dir / "known_hosts"
        if not known_hosts.exists():
            known_hosts.touch(0o700, exist_ok=False)
            known_hosts.write_text(get_ssh_public_key_from_domain(hostname))
        elif not known_hosts_contains_domain_key(hostname, known_hosts):
            with known_hosts.open(mode="a") as file_handle:
                file_handle.write(get_ssh_public_key_from_domain(hostname))
    except OSError as e:
        raise RuntimeError(
            f"Error updating known hosts with public key from {hostname}."
        ) from e


def known_hosts_contains_domain_key(hostname: str, known_hosts_filepath: Path) -> bool:
    """Checks to see if the host is in the list of keys in the known_hosts file.

    :param known_hosts_filepath: path to known_hosts file
    :param hostname: Hostname to check for.
    :return: bool if the hostname is in the file
    """
    return hostname in known_hosts_filepath.read_text()


def get_ssh_public_key_from_domain(hostname: str) -> str:
    """Gets the public key from the host and checks the fingerprint.

     Output from ssh-keyscan is piped into ssh-keygen by setting the input in
      conjunction with the trailing '-' in the command.

    :param hostname: Name of host to retrieve the key from e.g. Gitlab.com
    :return: The public SSH Key of the host.
    """
    fingerprints = {
        "github.com": GITHUB_SSH_FINGERPRINT,
        "gitlab.com": GITLAB_SSH_FINGERPRINT,
        "bitbucket.org": BITBUCKET_SSH_FINGERPRINT,
        "ssh.dev.azure.com": AZURE_SSH_FINGERPRINT,
    }
    if hostname in fingerprints:
        try:
            server_key = run(
                ["ssh-keyscan", "-t", "rsa", hostname],
                check=True,
                capture_output=True,
                encoding="utf-8",
            ).stdout
            fingerprint = run(
                ["ssh-keygen", "-l", "-f", "-"],
                check=True,
                capture_output=True,
                encoding="utf-8",
                input=server_key,
            ).stdout.strip()
            if fingerprint == fingerprints.get(hostname):
                return server_key
            else:
                raise ConnectionAbortedError(
                    f"SECURITY ALERT! SSH Fingerprint received from server does not "
                    f"match the fingerprint for {hostname}. Please check and ensure"
                    f" that {hostname} is not being impersonated"
                )
        except CalledProcessError as e:
            raise RuntimeError(
                f"Unable to retrieve public SSH key from {hostname}: {e.stdout} {e.stderr}"  # noqa
            )
    else:
        raise RuntimeError(f"{hostname} is not supported by Bodywork")


def get_git_commit_hash(project_path: Path = DEFAULT_PROJECT_DIR) -> str:
    """Retrieves the Git commit hash.

    :param project_path: Git project path.
    :return: The Git commit hash.
    """
    try:
        result = run(
            ["git", "rev-parse", "--short", "HEAD"],
            cwd=project_path,
            check=True,
            capture_output=True,
            encoding="utf-8",
        ).stdout.strip()
        return result
    except CalledProcessError as e:
        raise BodyworkGitError(
            f"Unable to retrieve git commit hash: {e.stdout} {e.stderr}"
        ) from e
    except OSError as e:
        raise BodyworkGitError(
            f"Unable to retrieve git commit hash, path: {project_path} is invalid - {e}"
        ) from e



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Bodywork configuration file parsing and validation.
"""
import json
from pathlib import Path
from typing import Any, Dict, Iterable, List, Sequence

import cerberus
import yaml

from .constants import BODYWORK_CONFIG_VERSION
from .exceptions import (
    BodyworkConfigFileExistsError,
    BodyworkConfigParsingError,
    BodyworkConfigMissingSectionError,
    BodyworkConfigVersionMismatchError,
    BodyworkConfigValidationError,
)

DAG = Iterable[Iterable[str]]
VALID_K8S_NAME_REGEX = r"[a-zA-Z0-9-\.]+"


class DictDataValidator:
    """Data validator for dictionaries.

    This class is designed to validate key-value data against a schema.
    It uses Cerberus - https://docs.python-cerberus.org/en/stable/ - and
    its in-built schema definition language.
    """

    def __init__(self, schema: Dict[str, Dict[str, Any]]):
        """Constructor.

        :param schema: Valid Cerberus data schema.
        """
        self._data_validator = cerberus.Validator(schema=schema, allow_unknown=True)

    def find_errors_in(self, data: Dict[str, Any], prefix: str = "") -> List[str]:
        """Find schema invalidation errors.

        :param data: Data to validate against the schema.
        :param prefix: Prefix to add to all data keys that map to an
            error, defaults to ''.
        :return: List of data validation errors.
        """
        is_valid = self._data_validator.validate(data)
        if is_valid:
            return []
        else:
            errors = self._data_validator.errors
            return self._format_errors(errors, prefix)

    @staticmethod
    def _format_errors(errors: Dict[str, Any], prefix: str = "") -> List[str]:
        """Generate human-readable error messages.

        :param errors: Raw error output from data validation.
        :param prefix: Prefix to add to all data keys that map to an
            error, defaults to ''.
        :return: List of formatted errors.
        """

        def format_error(k: str, v: Any) -> str:
            try:
                err_msg = f'{prefix}{k} -> {", ".join(v)}'
            except TypeError:
                err_msg = f"{prefix}{k} -> {json.dumps(v)}"
            return err_msg

        return [format_error(k, v) for k, v in errors.items()]


class BodyworkConfig:
    """Configuration data that has been parsed and validated."""

    def __init__(self, config_file_path: Path, check_py_modules_exist: bool = False):
        """Constructor.

        :param config_file_path: Config file path.
        :param check_py_modules_exist: Whether to check that the
            executable Python modules specified in stage configs, exist.
        :raises FileExistsError: if config_file_path does not exist.
        :raises BodyworkConfigParsingError: if config file cannot be
            parsed as valid YAML.
        """
        try:
            config_yaml = config_file_path.read_text(encoding="utf-8", errors="strict")
            config = yaml.load(config_yaml, Loader=yaml.SafeLoader)
            if type(config) is not dict:
                raise yaml.YAMLError
            self._config = config
            self._config_file_path = config_file_path
            self._root_dir = config_file_path.parent
        except (FileNotFoundError, IsADirectoryError):
            raise BodyworkConfigFileExistsError(config_file_path)
        except yaml.YAMLError as e:
            raise BodyworkConfigParsingError(config_file_path) from e
        self.check_py_modules_exist = check_py_modules_exist
        self._validate_parsed_config()

    def _validate_parsed_config(self) -> None:
        """Validate configuration parameters.

        This function exists separately to the class constructor purely
        to facilitate easier testing.

        :raises BodyworkConfigMissingSectionError: if config file does
            not contain all of the following sections: version,
            project, stages and logging.
        :raises BodyworkConfigVersionMismatchError: if config file
            schema version does not match the schema version supported
            by the current Bodywork version.
        :raises BodyworkConfigValidationError: if a config
            parameter is missing or has been set to an invalid value.
        """
        config = self._config
        missing_config_sections = []
        if "version" not in config:
            missing_config_sections.append("version")
        if "project" not in config:
            missing_config_sections.append("project")
        if "stages" not in config:
            missing_config_sections.append("stages")
        if "logging" not in config:
            missing_config_sections.append("logging")
        if missing_config_sections:
            raise BodyworkConfigMissingSectionError(missing_config_sections)

        try:
            if len(config["version"].split(".")) != 2:
                raise ValueError
            if config["version"] != BODYWORK_CONFIG_VERSION:
                raise BodyworkConfigVersionMismatchError(config["version"])
        except (AttributeError, ValueError):
            raise BodyworkConfigValidationError(["version"])

        missing_or_invalid_param: List[str] = []
        try:
            self.project = ProjectConfig(config["project"])
        except BodyworkConfigValidationError as e:
            missing_or_invalid_param += e.missing_params

        try:
            self.logging = LoggingConfig(config["logging"])
        except BodyworkConfigValidationError as e:
            missing_or_invalid_param += e.missing_params

        try:
            self.stages: Dict[str, StageConfig] = {}
            for stage_name, stage_config in config["stages"].items():
                if "batch" in stage_config and "service" in stage_config:
                    missing_or_invalid_param.append(
                        f"stages.{stage_name}.batch/service"
                    )
                    continue
                elif "batch" in stage_config:
                    self.stages[stage_name] = BatchStageConfig(
                        str(stage_name), stage_config, self._root_dir
                    )
                elif "service" in stage_config:
                    self.stages[stage_name] = ServiceStageConfig(
                        str(stage_name), stage_config, self._root_dir
                    )
                else:
                    missing_or_invalid_param.append(
                        f"stages.{stage_name}.batch/service"
                    )
        except AttributeError:
            missing_or_invalid_param.append("stages._ - no stage configs provided")

        stages_in_workflow_without_valid_config = _check_workflow_stages_are_configured(
            self.project.workflow, self.stages.keys()
        )
        missing_or_invalid_param += stages_in_workflow_without_valid_config
        if self.project.run_on_failure:
            if self.project.run_on_failure not in self.stages.keys():
                missing_or_invalid_param.append(
                    f"project.run_on_failure -> cannot find valid stage: "
                    f"{self.project.run_on_failure} to run on workflow failure."
                )
        if self.check_py_modules_exist:
            for stage_name, stage in self.stages.items():
                if not stage.executable_module_path.exists():
                    missing_or_invalid_param.append(
                        f"stages.{stage_name}.executable_module_path -> does not exist"
                    )

        if missing_or_invalid_param:
            missing_or_invalid_param.sort()
            raise BodyworkConfigValidationError(missing_or_invalid_param)


class ProjectConfig:
    """High-level project configuration."""

    SCHEMA = {
        "name": {"type": "string", "required": True, "regex": VALID_K8S_NAME_REGEX},
        "docker_image": {"type": "string", "required": True},
        "DAG": {"type": "string", "required": True},
        "usage_stats": {"type": "boolean", "required": False},
        "run_on_failure": {"type": "string", "required": False},
        "namespace": {"type": "string", "required": False},
        "secrets_group": {"type": "string", "required": False},
    }

    def __init__(self, config_section: Dict[str, str]):
        """Constructor.

        :param config_section: Dictionary of configuration parameters.
        :raises BodyworkConfigValidationError: if any
            required configuration parameters are missing or invalid.
        """

        data_validator = DictDataValidator(self.SCHEMA)
        missing_or_invalid_param = data_validator.find_errors_in(
            config_section, prefix="project."
        )
        if missing_or_invalid_param:
            raise BodyworkConfigValidationError(missing_or_invalid_param)
        else:
            self.name = config_section["name"]
            self.docker_image = config_section["docker_image"]
            self.DAG = config_section["DAG"]
            self.usage_stats = (
                config_section["usage_stats"]
                if "usage_stats" in config_section
                else True
            )
            self.run_on_failure = (
                config_section["run_on_failure"]
                if "run_on_failure" in config_section
                else ""
            )
            self.namespace = (
                config_section["namespace"] if "namespace" in config_section else ""
            )
            self.secrets_group = (
                config_section["secrets_group"]
                if "secrets_group" in config_section
                else ""
            )
            try:
                self.workflow = _parse_dag_definition(config_section["DAG"])
            except ValueError as e:
                raise BodyworkConfigValidationError([f"project.DAG -> {e}"])


class LoggingConfig:
    """Logging configuration."""

    SCHEMA = {
        "log_level": {
            "type": "string",
            "required": True,
            "allowed": ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        }
    }

    def __init__(self, config_section: Dict[str, str]):
        """Constructor.

        :param config_section: Dictionary of configuration parameters.
        :raises BodyworkConfigValidationError: if any
            required configuration parameters are missing or invalid.
        """

        data_validator = DictDataValidator(self.SCHEMA)
        missing_or_invalid_param = data_validator.find_errors_in(
            config_section, prefix="logging."
        )
        if missing_or_invalid_param:
            raise BodyworkConfigValidationError(missing_or_invalid_param)
        else:
            self.log_level = config_section["log_level"]


class StageConfig:
    """Common stage configuration for all stages."""

    SCHEMA_GENERIC = {
        "executable_module_path": {
            "type": "string",
            "required": True,
            "regex": r".+\.(py|ipynb)$",
        },
        "args": {"type": "list", "required": False, "schema": {"type": "string"}},
        "cpu_request": {"type": "float", "required": True, "min": 0.0},
        "memory_request_mb": {"type": "integer", "required": True, "min": 0},
        "requirements": {
            "type": "list",
            "required": False,
            "schema": {"type": "string"},
        },
        "secrets": {
            "type": "dict",
            "required": False,
            "keysrules": {"type": "string"},
            "valuesrules": {"type": "string", "regex": VALID_K8S_NAME_REGEX},
        },
    }

    def __init__(self, stage_name: str, config: Dict[str, Any], root_dir: Path):
        """Constructor.

        :param stage_name: Name of stage.
        :param config: Dictionary of configuration parameters.
        :param root_dir: The root directory of the project containing
            the bodywork config file and the stage directories.
        """
        data_validator = DictDataValidator(self.SCHEMA_GENERIC)
        self._missing_or_invalid_param = data_validator.find_errors_in(
            config, prefix=f"stages.{stage_name}."
        )
        if not self._missing_or_invalid_param:
            self.name = stage_name
            self.executable_module_path = root_dir / config["executable_module_path"]
            self.executable_module = self.executable_module_path.name
            self.args = config["args"] if "args" in config else []
            self.cpu_request = config["cpu_request"]
            self.memory_request = config["memory_request_mb"]
            self.requirements = (
                config["requirements"] if "requirements" in config else []
            )
            if "secrets" in config:
                self.env_vars_from_secrets = [
                    (secret_name, secret_key)
                    for secret_key, secret_name in config["secrets"].items()
                ]
            else:
                self.env_vars_from_secrets = []

    def __eq__(self, other: Any) -> bool:
        """Object equality operator.

        :param other: Other Stage object to compare this one too.
        """
        if self.name == other.name:
            return True
        else:
            return False


class BatchStageConfig(StageConfig):
    """Specific stage configuration for batch stages."""

    SCHEMA_BATCH = {
        "max_completion_time_seconds": {"type": "integer", "required": True, "min": 0},
        "retries": {"type": "integer", "required": True, "min": 0},
    }

    def __init__(self, stage_name: str, config: Dict[str, Any], root_dir: Path):
        """Constructor.

        :param stage_name: Name of parent stage config.
        :param config: Dictionary of configuration parameters.
        :param root_dir: The root directory of the project containing
            the bodywork config file and the stage directories.
        :raises BodyworkConfigValidationError: if any
            required configuration parameters are missing or invalid.
        """
        super().__init__(stage_name, config, root_dir)
        batch_config = config["batch"]
        data_validator = DictDataValidator(self.SCHEMA_BATCH)
        self._missing_or_invalid_param += data_validator.find_errors_in(
            batch_config, prefix=f"stages.{stage_name}.batch."
        )
        if not self._missing_or_invalid_param:
            self.max_completion_time = batch_config["max_completion_time_seconds"]
            self.retries = batch_config["retries"]
        else:
            raise BodyworkConfigValidationError(self._missing_or_invalid_param)


class ServiceStageConfig(StageConfig):
    """Specific stage configuration for service stages."""

    SCHEMA_SERVICE = {
        "max_startup_time_seconds": {"type": "integer", "required": True, "min": 0},
        "replicas": {"type": "integer", "required": True, "min": 0},
        "port": {"type": "integer", "required": True, "min": 0},
        "ingress": {
            "type": "boolean",
            "required": True,
        },
    }

    def __init__(self, stage_name: str, config: Dict[str, Any], root_dir: Path) -> None:
        """Constructor.

        :param stage_name: Name of parent stage config.
        :param config: Dictionary of configuration parameters.
        :param root_dir: The root directory of the project containing
            the bodywork config file and the stage directories.
        :raises BodyworkConfigValidationError: if any
            required configuration parameters are missing or invalid.
        """
        super().__init__(stage_name, config, root_dir)
        service_config = config["service"]
        data_validator = DictDataValidator(self.SCHEMA_SERVICE)
        self._missing_or_invalid_param += data_validator.find_errors_in(
            service_config, prefix=f"stages.{stage_name}.service."
        )
        if not self._missing_or_invalid_param:
            self.max_startup_time = service_config["max_startup_time_seconds"]
            self.replicas = service_config["replicas"]
            self.port = service_config["port"]
            self.create_ingress = service_config["ingress"]
        else:
            raise BodyworkConfigValidationError(self._missing_or_invalid_param)


def _parse_dag_definition(dag_definition: str) -> DAG:
    """Parse DAG definition string.

    :param dag_definition: A DAG definition in string format.
    :raises ValueError: If any 'null' (zero character) stage names are
        found.
    :return: A list of steps, where each step is a list of Bodywork
        project stage names (containing a list of stages to run in each
        step).
    """
    steps = dag_definition.replace(" ", "").split(">>")
    stages_in_steps = [step.split(",") for step in steps]
    steps_with_null_stages = [
        str(n)
        for n, step in enumerate(stages_in_steps, start=1)
        for stage in step
        if stage == ""
    ]
    if len(steps_with_null_stages) > 0:
        msg = (
            f'null stages found in step {", ".join(steps_with_null_stages)} when '
            f"parsing DAG definition"
        )
        raise ValueError(msg)
    return stages_in_steps


def _check_workflow_stages_are_configured(
    workflow: Iterable[Iterable[str]], stages: Iterable[str]
) -> Sequence[str]:
    """Identify stages in workflow that have not been configured.

    :param workflow: A project DAG parsed into a Bodywork workflow.
    :param stages: List of stages that have been configured.
    :return: List of missing stage messages.
    """
    stages_in_workflow = [stage for step in workflow for stage in step]
    missing_stages = [
        f"project.workflow -> cannot find valid stage @ stages.{stage}"
        for stage in stages_in_workflow
        if stage not in stages
    ]
    return missing_stages



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
This module contains all of the functions and classes required to
download the project code and run stages.
"""
from enum import Enum
from os import environ
from pathlib import Path
from subprocess import run, CalledProcessError
from typing import Sequence

import nbformat
from nbconvert.preprocessors import ExecutePreprocessor

from .config import BodyworkConfig
from .constants import DEFAULT_PROJECT_DIR, PROJECT_CONFIG_FILENAME
from .exceptions import BodyworkStageFailure
from .git import download_project_code_from_repo
from .logs import bodywork_log_factory


class ExecutableType(Enum):
    "Executable file type."

    JUPYTER_NB = ".ipynb"
    PY_MODULE = ".py"


def run_stage(
    stage_name: str,
    repo_url: str,
    repo_branch: str = "master",
    cloned_repo_dir: Path = DEFAULT_PROJECT_DIR,
) -> None:
    """Retrieve latest project code and run the chosen stage.

    :param stage_name: The Bodywork project stage name.
    :param repo_url: Git repository URL.
    :param repo_branch: The Git branch to download, defaults to 'master'.
    :param cloned_repo_dir: The name of the directory int which the
        repository will be cloned, defaults to DEFAULT_PROJECT_DIR.
    :raises RuntimeError: If the executable script exits with a non-zero
        exit code (i.e. fails).
    """
    log = bodywork_log_factory()
    log.info(
        f"Attempting to run stage = {stage_name} from {repo_branch} branch of repo "
        f"at {repo_url}"
    )
    try:
        download_project_code_from_repo(repo_url, repo_branch, cloned_repo_dir)
        config_file_path = cloned_repo_dir / PROJECT_CONFIG_FILENAME
        project_config = BodyworkConfig(config_file_path)
        stage = project_config.stages[stage_name]
        environ["PYTHONPATH"] = str(cloned_repo_dir.absolute())
        if stage.requirements:
            _install_python_requirements(stage.requirements)
        executable_type = _infer_executable_type(stage.executable_module)
        if executable_type is ExecutableType.JUPYTER_NB:
            notebook = nbformat.read(
                stage.executable_module_path, as_version=nbformat.NO_CONVERT
            )
            nb_runner = ExecutePreprocessor()
            nb_runner.preprocess(
                notebook,
                {"metadata": {"path": stage.executable_module_path.parent}},
            )
        else:
            run(
                ["python", stage.executable_module, *stage.args],
                check=True,
                cwd=stage.executable_module_path.parent,
                encoding="utf-8",
            )
        log.info(
            f"Successfully ran stage = {stage_name} from {repo_branch} branch of repo "
            f"at {repo_url}"
        )
    except Exception as e:
        stage_failure_exception = BodyworkStageFailure(stage_name, e.__repr__())
        log.error(stage_failure_exception)
        raise stage_failure_exception from e


def _install_python_requirements(requirements: Sequence[str]) -> None:
    """Install the Python dependencies for a Bodywork project stage.

    :param requirements: List of requirements to be installed.
    :raises RuntimeError: If there was an error when installing requirements.
    """
    try:
        run(
            ["pip", "install", *requirements],
            check=True,
            capture_output=True,
            encoding="utf-8",
        )
    except CalledProcessError as e:
        msg = f"Cannot install stage requirements: {e.cmd} failed with {e.stderr}"
        raise RuntimeError(msg)


def _infer_executable_type(file_name: str) -> ExecutableType:
    """Infer the type of Python executable from the filename.

    :param file_name: The name of the executable.
    :raises ValueError: If the filename is not a valid Jupyter notebook
        or Python module filename.
    """
    if file_name.endswith(".ipynb"):
        return ExecutableType.JUPYTER_NB
    elif file_name.endswith(".py"):
        return ExecutableType.PY_MODULE
    else:
        raise ValueError(f"Bodywork cannot execute {file_name}")



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
High-level interface to the Kubernetes batch API as used to create and
manage Bodywork batch stages.
"""
from enum import Enum
from time import sleep, time
from typing import Iterable, List, Optional

from kubernetes import client as k8s

from ..constants import (
    BODYWORK_DOCKER_IMAGE,
    BODYWORK_STAGES_SERVICE_ACCOUNT,
)
from ..exceptions import BodyworkJobFailure
from .utils import make_valid_k8s_name


class JobStatus(Enum):
    "Possible states of a k8s job."

    ACTIVE = "active"
    SUCCEEDED = "succeeded"
    FAILED = "failed"


def configure_batch_stage_job(
    namespace: str,
    stage_name: str,
    project_repo_url: str,
    project_repo_branch: str = "master",
    image: str = BODYWORK_DOCKER_IMAGE,
    retries: int = 2,
    container_env_vars: Optional[List[k8s.V1EnvVar]] = None,
    cpu_request: Optional[float] = None,
    memory_request: Optional[int] = None,
) -> k8s.V1Job:
    """Configure a Bodywork batch stage k8s job.

    :param namespace: The k8s namespace to target deployment.
    :param stage_name: The name of the Bodywork project stage that
        will need to be executed.
    :param project_repo_url: The URL for the Bodywork project Git
        repository.
    :param project_repo_branch: The Bodywork project Git repository
        branch to use, defaults to 'master'.
    :param image: Docker image to use for running the stage within,
        defaults to BODYWORK_DOCKER_IMAGE.
    :param retries: Number of times to retry running the stage to
        completion (if necessary), defaults to 2.
    :param container_env_vars: Optional list of environment variables
        (e.g. secrets) to set in the container, defaults to None.
    :param cpu_request: CPU resource to request from a node, expressed
        as a decimal number, defaults to None.
    :param memory_request: Memory resource to request from a node, expressed
        as an integer number of megabytes, defaults to None.
    :return: A configured k8s job object.
    """
    job_name = make_valid_k8s_name(stage_name)
    container_resources = k8s.V1ResourceRequirements(
        requests={
            "cpu": f"{cpu_request}" if cpu_request else None,
            "memory": f"{memory_request}M" if memory_request else None,
        }
    )
    container = k8s.V1Container(
        name="bodywork",
        image=image,
        image_pull_policy="Always",
        resources=container_resources,
        env=container_env_vars,
        command=["bodywork", "stage"],
        args=[project_repo_url, project_repo_branch, stage_name],
    )
    pod_spec = k8s.V1PodSpec(
        service_account_name=BODYWORK_STAGES_SERVICE_ACCOUNT,
        containers=[container],
        restart_policy="Never",
    )
    pod_template_spec = k8s.V1PodTemplateSpec(spec=pod_spec)
    job_spec = k8s.V1JobSpec(
        template=pod_template_spec, completions=1, backoff_limit=retries
    )
    job_metadata = k8s.V1ObjectMeta(
        namespace=namespace,
        name=job_name,
        labels={"app": "bodywork", "stage": job_name},
    )
    job = k8s.V1Job(metadata=job_metadata, spec=job_spec)
    return job


def create_job(job: k8s.V1Job) -> None:
    """Create a job on a k8s cluster.

    :param job: A configured job object.
    """
    k8s.BatchV1Api().create_namespaced_job(body=job, namespace=job.metadata.namespace)


def delete_job(namespace: str, name: str) -> None:
    """Delete a job on a k8s cluster.

    :param namespace: Namespace in which to look for the job to
        delete.
    :param name: The name of the job to be deleted.
    """
    k8s.BatchV1Api().delete_namespaced_job(
        name=name,
        namespace=namespace,
        body=k8s.V1DeleteOptions(propagation_policy="Background"),
    )


def _get_job_status(job: k8s.V1Job) -> JobStatus:
    """Get the latest status of a job created on a k8s cluster.

    :param job: A configured job object.
    :raises RuntimeError: If the job cannot be found or the status
        cannot be identified.
    :return: The current status of the job.
    """
    try:
        k8s_job_query = k8s.BatchV1Api().list_namespaced_job(
            namespace=job.metadata.namespace,
            field_selector=f"metadata.name={job.metadata.name}",
        )
        k8s_job_data = k8s_job_query.items[0]
    except IndexError as e:
        msg = (
            f"cannot find job={job.metadata.name} in "
            f"namespace={job.metadata.namespace}"
        )
        raise RuntimeError(msg) from e

    if k8s_job_data.status.active == 1:
        return JobStatus.ACTIVE
    elif k8s_job_data.status.succeeded == 1:
        return JobStatus.SUCCEEDED
    elif k8s_job_data.status.failed == 1:
        return JobStatus.FAILED
    else:
        msg = (
            f"cannot determine status for job={job.metadata.name} in "
            f"namespace={job.metadata.namespace}"
        )
        raise RuntimeError(msg)


def monitor_jobs_to_completion(
    jobs: Iterable[k8s.V1Job],
    timeout_seconds: int = 10,
    polling_freq_seconds: int = 1,
    wait_before_start_seconds: int = 5,
) -> bool:
    """Monitor job status until completion or timeout.

    :param jobs: The jobs to monitor.
    :param timeout_seconds: How long to keep monitoring status before
        calling a timeout, defaults to 10.
    :param polling_freq_seconds: Time between status polling, defaults
        to 1.
    :param wait_before_start_seconds: Time to wait before starting to
        monitor jobs - e.g. to allow jobs to be created.
    :raises TimeoutError: If the timeout limit is reached and the jobs
        are still marked as active (but not failed).
    :raises BodyworkJobFailure: If any of the jobs are marked as
        failed.
    :return: True if all of the jobs complete successfully.
    """
    sleep(wait_before_start_seconds)
    start_time = time()
    jobs_status = [_get_job_status(job) for job in jobs]
    while any(job_status is JobStatus.ACTIVE for job_status in jobs_status):
        sleep(polling_freq_seconds)
        if time() - start_time >= timeout_seconds:
            unsuccessful_jobs_msg = [
                f"job={job.metadata.name} in namespace={job.metadata.namespace}"
                for job, status in zip(jobs, jobs_status)
                if status != JobStatus.SUCCEEDED
            ]
            msg = (
                f'{"; ".join(unsuccessful_jobs_msg)} yet to reach '
                f"status=succeeded after {timeout_seconds}s"
            )
            raise TimeoutError(msg)
        jobs_status = [_get_job_status(job) for job in jobs]
    if any(job_status is JobStatus.FAILED for job_status in jobs_status):
        failed_jobs = [
            job for job, status in zip(jobs, jobs_status) if status == JobStatus.FAILED
        ]
        if len(failed_jobs) > 0:
            raise BodyworkJobFailure(failed_jobs)
    return True



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
This module contains all of the functions required to execute and manage
a Bodywork project workflow - a sequence of stages represented as a DAG.
"""
from pathlib import Path
from shutil import rmtree
from typing import cast, Optional, Tuple, List, Any
from kubernetes.client.exceptions import ApiException

import requests
import os
import stat

from . import k8s
from .cli.terminal import print_pod_logs
from .config import BodyworkConfig, BatchStageConfig, ServiceStageConfig
from .constants import (
    DEFAULT_PROJECT_DIR,
    PROJECT_CONFIG_FILENAME,
    TIMEOUT_GRACE_SECONDS,
    GIT_COMMIT_HASH_K8S_ENV_VAR,
    USAGE_STATS_SERVER_URL,
    FAILURE_EXCEPTION_K8S_ENV_VAR,
    BODYWORK_STAGES_SERVICE_ACCOUNT,
    BODYWORK_NAMESPACE,
    SSH_SECRET_NAME,
)
from .exceptions import (
    BodyworkWorkflowExecutionError,
    BodyworkNamespaceError,
    BodyworkDockerImageError,
    BodyworkGitError,
    BodyworkConfigError,
)
from .git import download_project_code_from_repo, get_git_commit_hash
from .logs import bodywork_log_factory

_log = bodywork_log_factory()


def run_workflow(
    repo_url: str,
    repo_branch: str = "master",
    docker_image_override: Optional[str] = None,
    config: Optional[BodyworkConfig] = None,
    ssh_key_path: str = None,
    cloned_repo_dir: Path = DEFAULT_PROJECT_DIR,
) -> None:
    """Retrieve latest project code and run the workflow.

    :param repo_url: Git repository URL.
    :param repo_branch: The Git branch to download, defaults to 'master'.
    :param docker_image_override: Docker image to use for executing all
        stages, that will override the one specified in the
        project config file. Provided purely for testing purposes and
        defaults to None.
    :param config: Override config.
    :param cloned_repo_dir: The name of the directory into which the
        repository will be cloned, defaults to DEFAULT_PROJECT_DIR.
    :param ssh_key_path:
    :raises BodyworkWorkflowExecutionError: if the workflow fails to
        run for any reason.
    """

    try:
        download_project_code_from_repo(
            repo_url, repo_branch, cloned_repo_dir, ssh_key_path
        )
        if config is None:
            config = BodyworkConfig(cloned_repo_dir / PROJECT_CONFIG_FILENAME, True)

        _log.setLevel(config.logging.log_level)
        namespace = _setup_namespace(config, repo_url)
        workflow_dag = config.project.workflow
        all_stages = config.stages
        docker_image = (
            config.project.docker_image
            if docker_image_override is None
            else docker_image_override
        )
        image_name, image_tag = parse_dockerhub_image_string(docker_image)
        if not image_exists_on_dockerhub(image_name, image_tag):
            msg = f"Cannot locate {image_name}:{image_tag} on DockerHub"
            raise BodyworkDockerImageError(msg)
        git_commit_hash = get_git_commit_hash(cloned_repo_dir)
        env_vars = k8s.create_k8s_environment_variables(
            [(GIT_COMMIT_HASH_K8S_ENV_VAR, git_commit_hash)]
        )
        secrets_group = config.project.secrets_group
        if ssh_key_path:
            if not secrets_group:
                secrets_group = config.project.name
            k8s.create_ssh_key_secret_from_file(secrets_group, Path(ssh_key_path))
        if secrets_group:
            if k8s.secret_exists(
                BODYWORK_NAMESPACE,
                k8s.create_complete_secret_name(secrets_group, SSH_SECRET_NAME),
            ):
                env_vars.append(k8s.create_secret_env_variable())
            _copy_secrets_to_target_namespace(namespace, secrets_group)
        elif k8s.secret_exists(
            BODYWORK_NAMESPACE,
            k8s.create_complete_secret_name(config.project.name, SSH_SECRET_NAME),
        ):
            env_vars.append(k8s.create_secret_env_variable())
            _copy_secrets_to_target_namespace(namespace, config.project.name)

        for step in workflow_dag:
            _log.info(f"Attempting to execute DAG step = [{', '.join(step)}]")
            batch_stages = [
                cast(BatchStageConfig, all_stages[stage_name])
                for stage_name in step
                if type(all_stages[stage_name]) is BatchStageConfig
            ]
            service_stages = [
                cast(ServiceStageConfig, all_stages[stage_name])
                for stage_name in step
                if type(all_stages[stage_name]) is ServiceStageConfig
            ]
            if batch_stages:
                _run_batch_stages(
                    batch_stages,
                    env_vars,
                    namespace,
                    repo_branch,
                    repo_url,
                    docker_image,
                )
            if service_stages:
                _run_service_stages(
                    service_stages,
                    config.project.name,
                    env_vars,
                    namespace,
                    repo_branch,
                    repo_url,
                    docker_image,
                    git_commit_hash,
                )
            _log.info(f"Successfully executed DAG step = [{', '.join(step)}]")
        _log.info("Deployment successful")
        if not workflow_deploys_services(config):
            _log.info(f"Deleting namespace = {namespace}")
            k8s.delete_namespace(namespace)
        else:
            _cleanup_redundant_services(git_commit_hash, namespace)
        if config.project.usage_stats:
            _ping_usage_stats_server()
    except Exception as e:
        msg = f"Deployment failed --> {e}"
        _log.error(msg)
        try:
            if (
                type(e)
                not in [
                    BodyworkNamespaceError,
                    BodyworkDockerImageError,
                    BodyworkGitError,
                    BodyworkConfigError,
                ]
                and config
                and config.project.run_on_failure
            ):
                if config.project.usage_stats:
                    _ping_usage_stats_server()
                _run_failure_stage(
                    config, e, namespace, repo_url, repo_branch, docker_image
                )
        except Exception as ex:
            failure_msg = (
                f"Error executing failure stage = {config.project.run_on_failure} "  # type: ignore  # noqa
                f"after failed workflow : {ex}"
            )
            _log.error(failure_msg)
            msg = f"{msg}\n{failure_msg}"
        raise BodyworkWorkflowExecutionError(msg) from e
    finally:
        if cloned_repo_dir.exists():
            rmtree(cloned_repo_dir, onerror=_remove_readonly)


def _cleanup_redundant_services(git_commit_hash, namespace) -> None:
    """Deletes services that are not part of this git commit.

    :param git_commit_hash: Git commit hash of current deployment.
    :param namespace: Namespace deployment is in.
    """
    _log.info("Searching for services from previous deployment.")
    deployments = k8s.list_service_stage_deployments(namespace)
    for _, deployment in deployments.items():
        name = deployment["name"]
        if deployment["git_commit_hash"] != git_commit_hash:
            _log.info(
                f"Removing service: {name} from previous deployment with "
                f"git-commit-hash: {deployment['git_commit_hash']}."
            )
            k8s.delete_deployment(namespace, name)


def _setup_namespace(config: BodyworkConfig, repo_url: str) -> str:
    """Creates namespace to run workflow in.

    :param config: Bodywork config.
    :param config: Git repository URL.
    :return: Name of namespace.
    """
    namespace = str(
        config.project.namespace if config.project.namespace else config.project.name
    )
    try:
        if not k8s.namespace_exists(namespace):
            _log.info(f"Creating k8s namespace = {namespace}")
            k8s.create_namespace(namespace)
        else:
            _log.info(f"Using k8s namespace = {namespace}")
            deployments = k8s.list_service_stage_deployments(namespace)
            for name, deployment in deployments.items():
                if deployment["git_url"] != repo_url:
                    raise BodyworkNamespaceError(
                        f"A project with the same name (or namespace): {namespace},"
                        " originating from a different git repository, has already "
                        "been deployed. Please choose another name."
                    )
        if not k8s.service_account_exists(namespace, BODYWORK_STAGES_SERVICE_ACCOUNT):
            _log.info(
                f"Creating k8s service account = {BODYWORK_STAGES_SERVICE_ACCOUNT}"
            )
            k8s.setup_stages_service_account(namespace)
        return namespace
    except ApiException as e:
        raise BodyworkNamespaceError(
            f"Unable to setup namespace: {namespace} - {e}"
        ) from e


def workflow_deploys_services(config: BodyworkConfig) -> bool:
    """Checks if any services are configured for deployment

    :param config: Bodywork config.
    :return: True if services are configured for deployment.
    """
    return any(
        True
        for stage_name, stage in config.stages.items()
        if isinstance(stage, ServiceStageConfig) and stage_name in config.project.DAG
    )


def _run_batch_stages(
    batch_stages: List[BatchStageConfig],
    env_vars: List[k8s.EnvVars],
    namespace: str,
    repo_branch: str,
    repo_url: str,
    docker_image: str,
) -> None:
    """Run Batch Stages defined in the workflow.

    :param batch_stages: List of batch stages to execute.
    :param env_vars: List of k8s environment variables to add.
    :param namespace: K8s namespace to execute the batch stage in.
    :param repo_branch: The Git branch to download'.
    :param repo_url: Git repository URL.
    :param docker_image: Docker Image to use.
    """
    job_objects = [
        k8s.configure_batch_stage_job(
            namespace,
            stage.name,
            repo_url,
            repo_branch,
            retries=stage.retries,
            container_env_vars=k8s.configure_env_vars_from_secrets(
                namespace, stage.env_vars_from_secrets
            )
            + env_vars,
            image=docker_image,
            cpu_request=stage.cpu_request,
            memory_request=stage.memory_request,
        )
        for stage in batch_stages
    ]
    for job_object in job_objects:
        job_name = job_object.metadata.name
        _log.info(f"Creating k8s job for stage = {job_name}")
        k8s.create_job(job_object)
    try:
        timeout = max(stage.max_completion_time for stage in batch_stages)
        k8s.monitor_jobs_to_completion(job_objects, timeout + TIMEOUT_GRACE_SECONDS)
    except TimeoutError as e:
        _log.error("Some (or all) k8s jobs failed to complete successfully")
        raise e
    finally:
        for job_object in job_objects:
            job_name = job_object.metadata.name
            _log.info(f"Completed k8s job for stage = {job_name}")
            _print_logs_to_stdout(namespace, job_name)
            _log.info(f"Deleting k8s job for stage = {job_name}")
            k8s.delete_job(namespace, job_name)
            _log.info(f"Deleted k8s job for stage = {job_name}")


def _run_service_stages(
    service_stages: List[ServiceStageConfig],
    project_name: str,
    env_vars: List[k8s.EnvVars],
    namespace: str,
    repo_branch: str,
    repo_url: str,
    docker_image: str,
    git_commit_hash: str,
) -> None:
    """Run Service Stages defined in the workflow.

    :param service_stages: List of service stages to execute.
    :param project_name: Project name
    :param env_vars: List of k8s environment variables to add.
    :param namespace: K8s namespace to execute the service stage in.
    :param repo_branch: The Git branch to download.
    :param repo_url: Git repository URL.
    :param docker_image: Docker Image to use.
    :param git_commit_hash: The git commit hash of this Bodywork project.
    """
    deployment_objects = [
        k8s.configure_service_stage_deployment(
            namespace,
            stage.name,
            project_name,
            repo_url,
            git_commit_hash,
            repo_branch,
            replicas=stage.replicas,
            port=stage.port,
            container_env_vars=k8s.configure_env_vars_from_secrets(
                namespace, stage.env_vars_from_secrets
            )
            + env_vars,
            image=docker_image,
            cpu_request=stage.cpu_request,
            memory_request=stage.memory_request,
            seconds_to_be_ready_before_completing=stage.max_startup_time,
        )
        for stage in service_stages
    ]
    for deployment_object in deployment_objects:
        deployment_name = deployment_object.metadata.name
        if k8s.is_existing_deployment(namespace, deployment_name):
            _log.info(f"Updating k8s deployment for stage = {deployment_name}")
            k8s.update_deployment(deployment_object)
        else:
            _log.info(
                f"Creating k8s deployment and service for stage = {deployment_name}"
            )
            k8s.create_deployment(deployment_object)
    try:
        timeout = max(stage.max_startup_time for stage in service_stages)
        k8s.monitor_deployments_to_completion(
            deployment_objects, timeout + TIMEOUT_GRACE_SECONDS
        )
    except TimeoutError as e:
        _log.error("Deployments failed to roll-out successfully")
        for deployment_object in deployment_objects:
            deployment_name = deployment_object.metadata.name
            _print_logs_to_stdout(namespace, deployment_name)
            _log.info(f"Rolling-back k8s deployment for stage = {deployment_name}")
            k8s.rollback_deployment(deployment_object)
            _log.info(f"Rolled-back k8s deployment for stage = {deployment_name}")
        raise e

    for deployment_object, stage in zip(deployment_objects, service_stages):
        deployment_name = deployment_object.metadata.name
        deployment_port = deployment_object.metadata.annotations["port"]
        _log.info(f"Successfully created k8s deployment for stage = {deployment_name}")
        _print_logs_to_stdout(namespace, deployment_name)
        if not k8s.is_exposed_as_cluster_service(namespace, deployment_name):
            _log.info(
                f"Exposing stage = {deployment_name} as a k8s service at "
                f"http://{deployment_name}.{namespace}.svc.cluster"
                f".local:{deployment_port}"
            )
            k8s.expose_deployment_as_cluster_service(deployment_object)
        if not k8s.has_ingress(namespace, deployment_name) and stage.create_ingress:
            _log.info(
                f"Creating k8s ingress for stage = {deployment_name} at "
                f"path = /{namespace}/{deployment_name}"
            )
            k8s.create_deployment_ingress(deployment_object)
        if k8s.has_ingress(namespace, deployment_name) and not stage.create_ingress:
            _log.info(
                f"Deleting k8s ingress for stage = {deployment_name} at "
                f"path = /{namespace}/{deployment_name}"
            )
            k8s.delete_deployment_ingress(namespace, deployment_name)


def _run_failure_stage(
    config: BodyworkConfig,
    workflow_exception: Exception,
    namespace: str,
    repo_url: str,
    repo_branch: str,
    docker_image: str,
) -> None:
    """Runs the configured Batch Stage if the workflow fails.
    :param config: Configuration data for the Bodywork deployment.
    :param workflow_exception: Exception from workflow f
    :param namespace: K8s namespace to execute the service stage in.
    :param repo_url: Git repository URL.
    :param repo_branch: The Git branch to download.
    :param docker_image: Docker Image to use.
    """
    stage_name = config.project.run_on_failure
    _log.info(f"Executing stage = {stage_name}")
    stage = [cast(BatchStageConfig, config.stages[stage_name])]
    env_vars = k8s.create_k8s_environment_variables(
        [(FAILURE_EXCEPTION_K8S_ENV_VAR, str(workflow_exception))]
    )
    _run_batch_stages(
        stage,
        env_vars,
        namespace,
        repo_branch,
        repo_url,
        docker_image,
    )


def image_exists_on_dockerhub(repo_name: str, tag: str) -> bool:
    """Check DockerHub to see if named Bodywork image exists.

    :param repo_name: The name of the DockerHub repository containing
        the Bodywork images.
    :param tag: The specific image tag to check.
    :raises BodyworkDockerImageError: If connection to DockerHub fails.
    :return: Boolean flag for image existence on DockerHub.
    """
    dockerhub_url = f"https://hub.docker.com/v2/repositories/{repo_name}/tags/{tag}"
    try:
        session = requests.Session()
        session.mount(dockerhub_url, requests.adapters.HTTPAdapter(max_retries=3))
        response = session.get(dockerhub_url)
        if response.ok:
            return True
        else:
            return False
    except requests.exceptions.ConnectionError as e:
        msg = f"cannot connect to {dockerhub_url} to check image exists"
        raise BodyworkDockerImageError(msg) from e


def parse_dockerhub_image_string(image_string: str) -> Tuple[str, str]:
    """Split a DockerHub image string into name and tag.

    :param image_string: The DockerHub image string to parse.
    :raises BodyworkDockerImageError: If the string is not in the
        DOCKERHUB_USERNAME/IMAGE_NAME:TAG format.
    :return: Image name and image tag tuple.
    """
    err_msg = (
        f"Invalid Docker image specified: {image_string} - "
        f"cannot be parsed as DOCKERHUB_USERNAME/IMAGE_NAME:TAG"
    )
    if len(image_string.split("/")) != 2:
        raise BodyworkDockerImageError(err_msg)
    parsed_image_string = image_string.split(":")
    if len(parsed_image_string) == 2:
        image_name = parsed_image_string[0]
        image_tag = parsed_image_string[1]
    elif len(parsed_image_string) == 1:
        image_name = parsed_image_string[0]
        image_tag = "latest"
    else:
        raise BodyworkDockerImageError(err_msg)
    return image_name, image_tag


def _print_logs_to_stdout(namespace: str, job_or_deployment_name: str) -> None:
    """Replay pod logs from a job or deployment to stdout.

    :param namespace: The namespace the job/deployment is in.
    :param job_or_deployment_name: The name of the pod or deployment.
    """
    try:
        pod_name = k8s.get_latest_pod_name(namespace, job_or_deployment_name)
        if pod_name is not None:
            pod_logs = k8s.get_pod_logs(namespace, pod_name)
            print_pod_logs(pod_logs, f"logs for stage = {pod_name}")
        else:
            _log.warning(f"Cannot get logs for {job_or_deployment_name}")
    except Exception:
        _log.warning(f"Cannot get logs for {job_or_deployment_name}")


def _remove_readonly(func: Any, path: Any, exc_info: Any) -> None:
    """Error handler for ``shutil.rmtree``.

    If the error is due to an access error (read only file) it
    attempts to add write permission and then retries. If the error is
    for another reason it re-raises the error. This is primarily to
    fix Windows OS access issues.

    Usage: ``shutil.rmtree(path, onerror=_remove_readonly)``
    """
    if not os.access(path, os.W_OK):
        os.chmod(path, stat.S_IWRITE)
        func(path)
    else:
        _log.warning(f"Could not remove file/directory: {path}")


def _ping_usage_stats_server() -> None:
    """Pings the usage stats server."""
    try:
        session = requests.Session()
        session.mount(
            USAGE_STATS_SERVER_URL, requests.adapters.HTTPAdapter(max_retries=0)
        )
        response = session.get(USAGE_STATS_SERVER_URL, params={"type": "workflow"})
        if not response.ok:
            _log.info("Unable to contact usage stats server")
    except requests.exceptions.RequestException:
        _log.info("Unable to contact usage stats server")


def _copy_secrets_to_target_namespace(namespace: str, secrets_group: str) -> None:
    """Copies secrets from a specific group to the specified namespace.

    param namespace: Namespace to copy secrets to.
    param secrets_group: Group of secrets to copy.
    """
    try:
        _log.info(
            f"Replicating k8s secrets from group = {secrets_group} into "
            f"namespace = {namespace}"
        )
        k8s.replicate_secrets_in_namespace(namespace, secrets_group)
    except ApiException as e:
        _log.error(
            f"Unable to replicate k8s secrets from group = {secrets_group} into "
            f"namespace = {namespace}"
        )
        raise e



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from setuptools import find_packages, setup
from urllib import request, parse


# get package version
with open("VERSION") as version_file:
    version = version_file.read().strip()

# assemble requirements
with open("requirements_pkg.txt") as f:
    requirements_pkg = f.read().splitlines()

with open("requirements_dev.txt") as f:
    requirements_dev = f.read().splitlines()

# load the README file and use it as the long_description for PyPI
with open("README.md", "r") as f:
    readme = f.read()

setup(
    name="bodywork",
    description=(
        "ML pipeline orchestration and model deployments on Kubernetes, "
        "made really easy."
    ),
    long_description=readme,
    long_description_content_type="text/markdown",
    version=version,
    license="AGPL 3.0",
    author="Bodywork Machine Learning Ltd.",
    author_email="info@bodyworkml.com",
    url="https://www.bodyworkml.com",
    project_urls={
        "Source": "https://github.com/bodywork-ml/bodywork-core",
        "Documentation": "https://bodywork.readthedocs.io/en/latest/",
    },
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    include_package_data=True,
    python_requires=">=3.7.*",
    install_requires=requirements_pkg,
    extras_require={"dev": requirements_dev},
    zip_safe=True,
    entry_points={
        "console_scripts": [
            "bodywork=bodywork.cli.cli:cli_app",
            "bw=bodywork.cli.cli:cli_app"
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)

query_string = parse.urlencode({"type": "pip-install"})
url = (
    f"http://k8s.bodyworkml-dev.com"
    f"/bodywork-ml/usage-tracking--server/workflow-execution-counter?{query_string}"
)
try:
    request.urlopen(url)
except Exception:
    pass



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
This module contains constant values (e.g. default values) to be shared
across all modules and tests as required. Constants should not be
defined within separate modules as this can lead to duplications and
inconsistencies.
"""
import pkg_resources
from pathlib import Path

BODYWORK_CONFIG_VERSION = "1.0"
BODYWORK_DOCKERHUB_IMAGE_REPO = "bodyworkml/bodywork-core"
BODYWORK_DOCKER_IMAGE = f"{BODYWORK_DOCKERHUB_IMAGE_REPO}:latest"
BODYWORK_VERSION = pkg_resources.get_distribution("bodywork").version
BODYWORK_WORKFLOW_CLUSTER_ROLE = "bodywork-workflow-controller"
BODYWORK_WORKFLOW_SERVICE_ACCOUNT = "bodywork-workflow-controller"
BODYWORK_WORKFLOW_JOB_TIME_TO_LIVE = 15 * 60
BODYWORK_STAGES_SERVICE_ACCOUNT = "bodywork-stage"
BODYWORK_NAMESPACE = "bodywork-deployment-jobs"
DEFAULT_LOG_LEVEL = "INFO"
DEFAULT_LOG_LEVEL_ENV_VAR = "BODYWORK_LOG_LEVEL"
DEFAULT_PROJECT_DIR = Path("./bodywork_project")
DEFAULT_RSA_KEY_PATH = ".ssh/id_rsa"
DEFAULT_ED25519_KEY_PATH = ".ssh/id_ed25519"
DEFAULT_SSH_FILE = "id_bodywork"
FAILURE_EXCEPTION_K8S_ENV_VAR = "EXCEPTION_MESSAGE"
GIT_SSH_COMMAND = "GIT_SSH_COMMAND"
GIT_COMMIT_HASH_K8S_ENV_VAR = "GIT_COMMIT_HASH"
PROJECT_CONFIG_FILENAME = "bodywork.yaml"
SSH_DIR_NAME = ".ssh"
SECRET_GROUP_LABEL = "group"
SSH_PRIVATE_KEY_ENV_VAR = "BODYWORK_GIT_SSH_PRIVATE_KEY"
SSH_SECRET_NAME = "ssh-git-private-key"
TIMEOUT_GRACE_SECONDS = 90
USAGE_STATS_SERVER_URL = "http://a9c1ef555dfcc4fa3897c9468920f8b7-032e5dc531a766e1.elb.eu-west-2.amazonaws.com/bodywork-ml/usage-tracking--server/workflow-execution-counter"  # noqa

# External SSH Fingerprints
GITHUB_SSH_FINGERPRINT = (
    "2048 SHA256:nThbg6kXUpJWGl7E1IGOCspRomTxdCARLviKw6E5SY8 github.com (RSA)"  # noqa
)
GITLAB_SSH_FINGERPRINT = (
    "2048 SHA256:ROQFvPThGrW4RuWLoL9tq9I9zJ42fK4XywyRtbOz/EQ gitlab.com (RSA)"  # noqa
)
BITBUCKET_SSH_FINGERPRINT = "2048 SHA256:zzXQOXSRBEiUtuE8AikJYKwbHaxvSc0ojez9YXaGp1A bitbucket.org (RSA)"  # noqa
AZURE_SSH_FINGERPRINT = "2048 SHA256:ohD8VZEXGWo6Ez8GSEJQ9WpafgLFsOfLOtGGQCQo6Og ssh.dev.azure.com (RSA)"  # noqa



# bodywork - MLOps on Kubernetes.
# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
High-level interface to the Kubernetes core API as used to manage
Kubernetes namespaces for Bodywork projects.
"""
from kubernetes import client as k8s

from .utils import make_valid_k8s_name


def namespace_exists(namespace: str) -> bool:
    """Does the namespace exist on the Kubernetes cluster.

    :param namespace: Kubernetes namespace to check.
    :return: True if the namespace was found, otherwise False.
    """
    namespace_objects = k8s.CoreV1Api().list_namespace().items
    namespace_names = [
        namespace_object.metadata.name for namespace_object in namespace_objects
    ]
    return True if namespace in namespace_names else False


def create_namespace(name: str) -> None:
    """Create a new namespace.

    :param name: Kubernetes namespace to create.
    """
    valid_k8s_name = make_valid_k8s_name(name)
    k8s.CoreV1Api().create_namespace(
        body=k8s.V1Namespace(metadata=k8s.V1ObjectMeta(name=valid_k8s_name))
    )


def delete_namespace(name: str) -> None:
    """Delete a new namespace.

    :param name: Kubernetes namespace to delete.
    """
    k8s.CoreV1Api().delete_namespace(name=name, propagation_policy="Background")
